# -*- coding: utf-8 -*-

"""
* Name:         interactive-clustering/tests/clustering/test_hierarchical.py
* Description:  Unittests for the `clustering.hierarchical` module.
* Author:       Erwan Schild
* Created:      17/03/2021
* Licence:      CeCILL (https://cecill.info/licences.fr.html)
"""

# ==============================================================================
# IMPORT PYTHON DEPENDENCIES
# ==============================================================================

import numpy as np
import pytest
from scipy.sparse import csr_matrix

from cognitivefactory.interactive_clustering.clustering.hierarchical import Cluster, HierarchicalConstrainedClustering
from cognitivefactory.interactive_clustering.constraints.binary import BinaryConstraintsManager


# ==============================================================================
# test_HierarchicalConstrainedClustering_for_inconsistent_linkage
# ==============================================================================
def test_HierarchicalConstrainedClustering_for_inconsistent_linkage():
    """
    Test that the `clustering.hierarchical.HierarchicalConstrainedClustering` initialization raises an `ValueError` for inconsistent `linkage` parameter.
    """

    # Check `ValueError` for bad string value for `linkage`.
    with pytest.raises(ValueError, match="`linkage`"):
        HierarchicalConstrainedClustering(
            linkage="as_you_want",
        )


# ==============================================================================
# test_HierarchicalConstrainedClustering_for_correct_settings
# ==============================================================================
def test_HierarchicalConstrainedClustering_for_correct_settings():
    """
    Test that the `clustering.hierarchical.HierarchicalConstrainedClustering` initialization runs correctly with the correct settings.
    """

    # Check a correct initialization.
    clustering_model = HierarchicalConstrainedClustering(
        linkage="average",
        random_seed=2,
    )
    assert clustering_model
    assert clustering_model.linkage == "average"
    assert clustering_model.random_seed == 2


# ==============================================================================
# test_HierarchicalConstrainedClustering_cluster_for_inconsistent_constraints_manager
# ==============================================================================
def test_HierarchicalConstrainedClustering_cluster_for_inconsistent_constraints_manager():
    """
    Test that the `clustering.spectral.HierarchicalConstrainedClustering` clustering raises an `ValueError` for inconsistent `constraints_manager` parameter.
    """

    # Initialize a `HierarchicalConstrainedClustering` instance.
    clustering_model = HierarchicalConstrainedClustering()

    # Check `ValueError` for not matrix `vectors`.
    with pytest.raises(ValueError, match="`constraints_manager`"):
        clustering_model.cluster(
            constraints_manager=None,
            vectors=None,
            nb_clusters=2,
        )


# ==============================================================================
# test_HierarchicalConstrainedClustering_cluster_for_inconsistent_vectors
# ==============================================================================
def test_HierarchicalConstrainedClustering_cluster_for_inconsistent_vectors():
    """
    Test that the `clustering.spectral.HierarchicalConstrainedClustering` clustering raises an `ValueError` for inconsistent `vectors` parameter.
    """

    # Initialize a `HierarchicalConstrainedClustering` instance.
    clustering_model = HierarchicalConstrainedClustering()

    # Check `ValueError` for not matrix `vectors`.
    with pytest.raises(ValueError, match="`vectors`"):
        clustering_model.cluster(
            constraints_manager=BinaryConstraintsManager(list_of_data_IDs=["first", "second", "third"]),
            vectors=None,
            nb_clusters=2,
        )


# ==============================================================================
# test_HierarchicalConstrainedClustering_cluster_for_inconsistent_nb_clusters
# ==============================================================================
def test_HierarchicalConstrainedClustering_cluster_for_inconsistent_nb_clusters():
    """
    Test that the `clustering.spectral.HierarchicalConstrainedClustering` clustering raises an `ValueError` for inconsistent `nb_clusters` parameter.
    """

    # Initialize a `HierarchicalConstrainedClustering` instance.
    clustering_model = HierarchicalConstrainedClustering()

    # Check `ValueError` for too small `nb_clusters`.
    with pytest.raises(ValueError, match="`nb_clusters`"):
        clustering_model.cluster(
            constraints_manager=BinaryConstraintsManager(list_of_data_IDs=["first", "second", "third"]),
            vectors={"first": np.array([1, 2, 3]), "second": np.array([[4, 5, 6]]), "third": csr_matrix([7, 8, 9])},
            nb_clusters=-1,
        )


# ==============================================================================
# test_HierarchicalConstrainedClustering_cluster_with_ward_linkage
# ==============================================================================
def test_HierarchicalConstrainedClustering_cluster_with_ward_linkage():
    """
    Test that the `clustering.hierarchical.HierarchicalConstrainedClustering` clustering works with ward `linkage`.
    """

    # Define `vectors` and `constraints_manager`
    vectors = {
        "0": csr_matrix([1.00, 0.00, 0.00]),
        "1": csr_matrix([0.95, 0.02, 0.01]),
        "2": csr_matrix([0.98, 0.00, 0.00]),
        "3": csr_matrix([0.99, 0.00, 0.00]),
        "4": csr_matrix([0.01, 0.99, 0.07]),
        "5": csr_matrix([0.02, 0.99, 0.07]),
        "6": csr_matrix([0.01, 0.99, 0.02]),
        "7": csr_matrix([0.01, 0.01, 0.97]),
        "8": csr_matrix([0.00, 0.01, 0.99]),
        "9": csr_matrix([0.00, 0.00, 1.00]),
    }
    constraints_manager = BinaryConstraintsManager(list_of_data_IDs=["0", "1", "2", "3", "4", "5", "6", "7", "8", "9"])
    constraints_manager.add_constraint(data_ID1="0", data_ID2="1", constraint_type="MUST_LINK")
    constraints_manager.add_constraint(data_ID1="5", data_ID2="7", constraint_type="CANNOT_LINK")

    # Initialize a `HierarchicalConstrainedClustering` instance.
    clustering_model = HierarchicalConstrainedClustering(
        linkage="ward",
        random_seed=1,
    )

    # Run clustering 3 clusters and some constraints.
    dict_of_predicted_clusters = clustering_model.cluster(
        constraints_manager=constraints_manager,
        vectors=vectors,
        nb_clusters=3,
    )

    assert clustering_model.dict_of_predicted_clusters
    assert dict_of_predicted_clusters == {
        "0": 0,
        "1": 0,
        "2": 0,
        "3": 0,
        "4": 1,
        "5": 1,
        "6": 1,
        "7": 2,
        "8": 2,
        "9": 2,
    }


# ==============================================================================
# test_HierarchicalConstrainedClustering_cluster_with_average_linkage
# ==============================================================================
def test_HierarchicalConstrainedClustering_cluster_with_average_linkage():
    """
    Test that the `clustering.hierarchical.HierarchicalConstrainedClustering` clustering works with average `linkage`.
    """

    # Define `vectors` and `constraints_manager`
    vectors = {
        "0": csr_matrix([1.00, 0.00, 0.00]),
        "1": csr_matrix([0.95, 0.02, 0.01]),
        "2": csr_matrix([0.98, 0.00, 0.00]),
        "3": csr_matrix([0.99, 0.00, 0.00]),
        "4": csr_matrix([0.01, 0.99, 0.07]),
        "5": csr_matrix([0.02, 0.99, 0.07]),
        "6": csr_matrix([0.01, 0.99, 0.02]),
        "7": csr_matrix([0.01, 0.01, 0.97]),
        "8": csr_matrix([0.00, 0.01, 0.99]),
        "9": csr_matrix([0.00, 0.00, 1.00]),
    }
    constraints_manager = BinaryConstraintsManager(list_of_data_IDs=["0", "1", "2", "3", "4", "5", "6", "7", "8", "9"])
    constraints_manager.add_constraint(data_ID1="0", data_ID2="1", constraint_type="MUST_LINK")
    constraints_manager.add_constraint(data_ID1="5", data_ID2="7", constraint_type="CANNOT_LINK")

    # Initialize a `HierarchicalConstrainedClustering` instance.
    clustering_model = HierarchicalConstrainedClustering(
        linkage="average",
        random_seed=1,
    )

    # Run clustering 3 clusters and some constraints.
    dict_of_predicted_clusters = clustering_model.cluster(
        constraints_manager=constraints_manager,
        vectors=vectors,
        nb_clusters=3,
    )

    assert clustering_model.dict_of_predicted_clusters
    assert dict_of_predicted_clusters == {
        "0": 0,
        "1": 0,
        "2": 0,
        "3": 0,
        "4": 1,
        "5": 1,
        "6": 1,
        "7": 2,
        "8": 2,
        "9": 2,
    }


# ==============================================================================
# test_HierarchicalConstrainedClustering_cluster_with_single_linkage
# ==============================================================================
def test_HierarchicalConstrainedClustering_cluster_with_single_linkage():
    """
    Test that the `clustering.hierarchical.HierarchicalConstrainedClustering` clustering works with single `linkage`.
    """

    # Define `vectors` and `constraints_manager`
    vectors = {
        "0": csr_matrix([1.00, 0.00, 0.00]),
        "1": csr_matrix([0.95, 0.02, 0.01]),
        "2": csr_matrix([0.98, 0.00, 0.00]),
        "3": csr_matrix([0.99, 0.00, 0.00]),
        "4": csr_matrix([0.01, 0.99, 0.07]),
        "5": csr_matrix([0.02, 0.99, 0.07]),
        "6": csr_matrix([0.01, 0.99, 0.02]),
        "7": csr_matrix([0.01, 0.01, 0.97]),
        "8": csr_matrix([0.00, 0.01, 0.99]),
        "9": csr_matrix([0.00, 0.00, 1.00]),
    }
    constraints_manager = BinaryConstraintsManager(list_of_data_IDs=["0", "1", "2", "3", "4", "5", "6", "7", "8", "9"])
    constraints_manager.add_constraint(data_ID1="0", data_ID2="1", constraint_type="MUST_LINK")
    constraints_manager.add_constraint(data_ID1="5", data_ID2="7", constraint_type="CANNOT_LINK")

    # Initialize a `HierarchicalConstrainedClustering` instance.
    clustering_model = HierarchicalConstrainedClustering(
        linkage="single",
        random_seed=1,
    )

    # Run clustering 3 clusters and some constraints.
    dict_of_predicted_clusters = clustering_model.cluster(
        constraints_manager=constraints_manager,
        vectors=vectors,
        nb_clusters=3,
    )

    assert clustering_model.dict_of_predicted_clusters
    assert dict_of_predicted_clusters == {
        "0": 0,
        "1": 0,
        "2": 0,
        "3": 0,
        "4": 1,
        "5": 1,
        "6": 1,
        "7": 2,
        "8": 2,
        "9": 2,
    }


# ==============================================================================
# test_HierarchicalConstrainedClustering_cluster_with_complete_linkage
# ==============================================================================
def test_HierarchicalConstrainedClustering_cluster_with_complete_linkage():
    """
    Test that the `clustering.hierarchical.HierarchicalConstrainedClustering` clustering works with complete `linkage`.
    """

    # Define `vectors` and `constraints_manager`
    vectors = {
        "0": csr_matrix([1.00, 0.00, 0.00]),
        "1": csr_matrix([0.95, 0.02, 0.01]),
        "2": csr_matrix([0.98, 0.00, 0.00]),
        "3": csr_matrix([0.99, 0.00, 0.00]),
        "4": csr_matrix([0.01, 0.99, 0.07]),
        "5": csr_matrix([0.02, 0.99, 0.07]),
        "6": csr_matrix([0.01, 0.99, 0.02]),
        "7": csr_matrix([0.01, 0.01, 0.97]),
        "8": csr_matrix([0.00, 0.01, 0.99]),
        "9": csr_matrix([0.00, 0.00, 1.00]),
    }
    constraints_manager = BinaryConstraintsManager(list_of_data_IDs=["0", "1", "2", "3", "4", "5", "6", "7", "8", "9"])
    constraints_manager.add_constraint(data_ID1="0", data_ID2="1", constraint_type="MUST_LINK")
    constraints_manager.add_constraint(data_ID1="5", data_ID2="7", constraint_type="CANNOT_LINK")

    # Initialize a `HierarchicalConstrainedClustering` instance.
    clustering_model = HierarchicalConstrainedClustering(
        linkage="complete",
        random_seed=1,
    )

    # Run clustering 3 clusters and some constraints.
    dict_of_predicted_clusters = clustering_model.cluster(
        constraints_manager=constraints_manager,
        vectors=vectors,
        nb_clusters=3,
    )

    assert clustering_model.dict_of_predicted_clusters
    assert dict_of_predicted_clusters == {
        "0": 0,
        "1": 0,
        "2": 0,
        "3": 0,
        "4": 1,
        "5": 1,
        "6": 1,
        "7": 2,
        "8": 2,
        "9": 2,
    }


# ==============================================================================
# test_HierarchicalConstrainedClustering_cluster_with_no_constraints_1
# ==============================================================================
def test_HierarchicalConstrainedClustering_cluster_with_no_constraints_1():
    """
    Test that the `clustering.hierarchical.HierarchicalConstrainedClustering` clustering works with no `constraints`.
    """

    # Define `vectors` and `constraints_manager`
    vectors = {
        "0": csr_matrix([1.00, 0.00, 0.00, 0.00]),
        "1": csr_matrix([0.95, 0.02, 0.02, 0.01]),
        "2": csr_matrix([0.98, 0.00, 0.02, 0.00]),
        "3": csr_matrix([0.99, 0.00, 0.01, 0.00]),
        "4": csr_matrix([0.50, 0.22, 0.21, 0.07]),
        "5": csr_matrix([0.50, 0.21, 0.22, 0.07]),
        "6": csr_matrix([0.01, 0.01, 0.01, 0.97]),
        "7": csr_matrix([0.00, 0.01, 0.00, 0.99]),
        "8": csr_matrix([0.00, 0.00, 0.00, 1.00]),
    }
    constraints_manager = BinaryConstraintsManager(list_of_data_IDs=list(vectors.keys()))

    # Initialize a `HierarchicalConstrainedClustering` instance.
    clustering_model = HierarchicalConstrainedClustering(
        random_seed=2,
    )

    # Run clustering 2 clusters and no constraints.
    dict_of_predicted_clusters = clustering_model.cluster(
        constraints_manager=constraints_manager,
        vectors=vectors,
        nb_clusters=2,
    )

    assert clustering_model.dict_of_predicted_clusters
    assert dict_of_predicted_clusters == {"0": 0, "1": 0, "2": 0, "3": 0, "4": 0, "5": 0, "6": 1, "7": 1, "8": 1}


# ==============================================================================
# test_HierarchicalConstrainedClustering_cluster_with_no_constraints_2
# ==============================================================================
def test_HierarchicalConstrainedClustering_cluster_with_no_constraints_2():
    """
    Test that the `clustering.hierarchical.HierarchicalConstrainedClustering` clustering works with no `constraints`.
    """

    # Define `vectors` and `constraints_manager`
    vectors = {
        "0": csr_matrix([1.00, 0.00, 0.00]),
        "1": csr_matrix([0.95, 0.02, 0.01]),
        "2": csr_matrix([0.98, 0.00, 0.00]),
        "3": csr_matrix([0.99, 0.00, 0.00]),
        "4": csr_matrix([0.01, 0.99, 0.07]),
        "5": csr_matrix([0.02, 0.99, 0.07]),
        "6": csr_matrix([0.01, 0.99, 0.02]),
        "7": csr_matrix([0.01, 0.01, 0.97]),
        "8": csr_matrix([0.00, 0.01, 0.99]),
        "9": csr_matrix([0.00, 0.00, 1.00]),
    }
    constraints_manager = BinaryConstraintsManager(list_of_data_IDs=list(vectors.keys()))

    # Initialize a `HierarchicalConstrainedClustering` instance.
    clustering_model = HierarchicalConstrainedClustering(
        random_seed=2,
    )

    # Run clustering 3 clusters and no constraints.
    dict_of_predicted_clusters = clustering_model.cluster(
        constraints_manager=constraints_manager,
        vectors=vectors,
        nb_clusters=3,
    )
    assert clustering_model.dict_of_predicted_clusters
    assert dict_of_predicted_clusters == {
        "0": 0,
        "1": 0,
        "2": 0,
        "3": 0,
        "4": 1,
        "5": 1,
        "6": 1,
        "7": 2,
        "8": 2,
        "9": 2,
    }


# ==============================================================================
# test_HierarchicalConstrainedClustering_cluster_with_some_constraints
# ==============================================================================
def test_HierarchicalConstrainedClustering_cluster_with_some_constraints():
    """
    Test that the `clustering.hierarchical.HierarchicalConstrainedClustering` clustering works with some `constraints`.
    """

    # Define `vectors` and `constraints_manager`
    vectors = {
        "0": csr_matrix([1.00, 0.00, 0.00, 0.00]),
        "1": csr_matrix([0.95, 0.02, 0.02, 0.01]),
        "2": csr_matrix([0.98, 0.00, 0.02, 0.00]),
        "3": csr_matrix([0.99, 0.00, 0.01, 0.00]),
        "4": csr_matrix([0.50, 0.22, 0.21, 0.07]),
        "5": csr_matrix([0.50, 0.21, 0.22, 0.07]),
        "6": csr_matrix([0.01, 0.01, 0.01, 0.97]),
        "7": csr_matrix([0.00, 0.01, 0.00, 0.99]),
        "8": csr_matrix([0.00, 0.00, 0.00, 1.00]),
    }
    constraints_manager = BinaryConstraintsManager(list_of_data_IDs=["0", "1", "2", "3", "4", "5", "6", "7", "8"])
    constraints_manager.add_constraint(data_ID1="0", data_ID2="1", constraint_type="MUST_LINK")
    constraints_manager.add_constraint(data_ID1="0", data_ID2="6", constraint_type="MUST_LINK")
    constraints_manager.add_constraint(data_ID1="0", data_ID2="7", constraint_type="MUST_LINK")
    constraints_manager.add_constraint(data_ID1="0", data_ID2="8", constraint_type="MUST_LINK")
    constraints_manager.add_constraint(data_ID1="4", data_ID2="5", constraint_type="MUST_LINK")
    constraints_manager.add_constraint(data_ID1="0", data_ID2="4", constraint_type="CANNOT_LINK")
    constraints_manager.add_constraint(data_ID1="2", data_ID2="4", constraint_type="CANNOT_LINK")

    # Initialize a `HierarchicalConstrainedClustering` instance.
    clustering_model = HierarchicalConstrainedClustering(
        random_seed=2,
    )

    # Run clustering 2 clusters and somme constraints.
    dict_of_predicted_clusters = clustering_model.cluster(
        constraints_manager=constraints_manager,
        vectors=vectors,
        nb_clusters=3,
    )
    assert clustering_model.dict_of_predicted_clusters
    assert dict_of_predicted_clusters == {
        "0": 0,
        "1": 0,
        "2": 1,
        "3": 1,
        "4": 2,
        "5": 2,
        "6": 0,
        "7": 0,
        "8": 0,
    }


# ==============================================================================
# test_HierarchicalConstrainedClustering_cluster_with_full_constraints
# ==============================================================================
def test_HierarchicalConstrainedClustering_cluster_with_full_constraints():
    """
    Test that the `clustering.hierarchical.HierarchicalConstrainedClustering` clustering works with full `constraints`.
    """

    # Define `vectors` and `constraints_manager`
    vectors = {
        "0": csr_matrix([1.00, 0.00, 0.00, 0.00]),
        "1": csr_matrix([0.95, 0.02, 0.02, 0.01]),
        "2": csr_matrix([0.98, 0.00, 0.02, 0.00]),
        "3": csr_matrix([0.99, 0.00, 0.01, 0.00]),
        "4": csr_matrix([0.50, 0.22, 0.21, 0.07]),
        "5": csr_matrix([0.50, 0.21, 0.22, 0.07]),
        "6": csr_matrix([0.01, 0.01, 0.01, 0.97]),
        "7": csr_matrix([0.00, 0.01, 0.00, 0.99]),
        "8": csr_matrix([0.00, 0.00, 0.00, 1.00]),
    }
    constraints_manager = BinaryConstraintsManager(list_of_data_IDs=["0", "1", "2", "3", "4", "5", "6", "7", "8"])
    constraints_manager.add_constraint(data_ID1="0", data_ID2="4", constraint_type="MUST_LINK")
    constraints_manager.add_constraint(data_ID1="0", data_ID2="8", constraint_type="MUST_LINK")
    constraints_manager.add_constraint(data_ID1="1", data_ID2="5", constraint_type="MUST_LINK")
    constraints_manager.add_constraint(data_ID1="2", data_ID2="6", constraint_type="MUST_LINK")
    constraints_manager.add_constraint(data_ID1="3", data_ID2="7", constraint_type="MUST_LINK")
    constraints_manager.add_constraint(data_ID1="0", data_ID2="1", constraint_type="CANNOT_LINK")
    constraints_manager.add_constraint(data_ID1="0", data_ID2="2", constraint_type="CANNOT_LINK")
    constraints_manager.add_constraint(data_ID1="0", data_ID2="3", constraint_type="CANNOT_LINK")
    constraints_manager.add_constraint(data_ID1="1", data_ID2="2", constraint_type="CANNOT_LINK")
    constraints_manager.add_constraint(data_ID1="1", data_ID2="3", constraint_type="CANNOT_LINK")
    constraints_manager.add_constraint(data_ID1="2", data_ID2="3", constraint_type="CANNOT_LINK")

    # Initialize a `HierarchicalConstrainedClustering` instance.
    clustering_model = HierarchicalConstrainedClustering()

    # Run clustering 4 clusters and full constraints.
    dict_of_predicted_clusters = clustering_model.cluster(
        constraints_manager=constraints_manager,
        vectors=vectors,
        nb_clusters=4,
    )
    assert clustering_model.dict_of_predicted_clusters
    assert dict_of_predicted_clusters == {
        "0": 0,
        "1": 1,
        "2": 2,
        "3": 3,
        "4": 0,
        "5": 1,
        "6": 2,
        "7": 3,
        "8": 0,
    }


# ==============================================================================
# test_HierarchicalConstrainedClustering_compute_predicted_clusters_without_clustering_tree
# ==============================================================================
def test_HierarchicalConstrainedClustering_compute_predicted_clusters_without_clustering_tree():
    """
    Test that the `compute_predicted_clusters` method of the `HierarchicalConstrainedClustering` raises `ValueError` if clustering is not run.
    """

    # Initialize a `HierarchicalConstrainedClustering` instance.
    clustering_model = HierarchicalConstrainedClustering(
        linkage="single",
        random_seed=1,
    )

    # Run `compute_predicted_clusters` without computing the clustering tree.
    with pytest.raises(ValueError, match="`clustering_root`"):
        clustering_model.compute_predicted_clusters(
            nb_clusters=2,
            by="size",
        )


# ==============================================================================
# test_HierarchicalConstrainedClustering_compute_predicted_clusters_travelling_by_size
# ==============================================================================
def test_HierarchicalConstrainedClustering_compute_predicted_clusters_travelling_by_size():
    """
    Test that the `compute_predicted_clusters` method of the `HierarchicalConstrainedClustering` clustering works by travelling by `"size"`.
    """

    # Define `vectors` and `constraints_manager`
    vectors = {
        "00": csr_matrix([1.00, 0.00, 0.00]),
        "01": csr_matrix([0.99, 0.00, 0.00]),
        "02": csr_matrix([0.97, 0.00, 0.00]),
        "03": csr_matrix([0.96, 0.00, 0.00]),
        "04": csr_matrix([0.94, 0.00, 0.00]),
        "05": csr_matrix([0.93, 0.00, 0.00]),
        "06": csr_matrix([0.80, 0.80, 0.00]),
        "07": csr_matrix([0.80, 0.81, 0.00]),
        "08": csr_matrix([0.00, 0.00, 0.70]),
        "09": csr_matrix([0.00, 0.00, 0.71]),
        "10": csr_matrix([0.00, 0.00, 0.99]),
        "11": csr_matrix([0.00, 0.00, 1.00]),
    }
    constraints_manager = BinaryConstraintsManager(list_of_data_IDs=list(vectors.keys()))

    # Initialize a `HierarchicalConstrainedClustering` instance.
    clustering_model = HierarchicalConstrainedClustering(
        linkage="single",
        random_seed=1,
    )

    # Compute all clustering tree.
    clustering_model.cluster(
        constraints_manager=constraints_manager,
        vectors=vectors,
        nb_clusters=2,
    )

    # Run `compute_predicted_clusters` while travalleing clustering tree by `"size"`.
    dict_of_predicted_clusters = clustering_model.compute_predicted_clusters(
        nb_clusters=4,
        by="size",
    )
    assert dict_of_predicted_clusters == {
        "00": 0,
        "01": 0,
        "02": 0,
        "03": 0,
        "04": 1,
        "05": 1,
        "06": 2,
        "07": 2,
        "08": 3,
        "09": 3,
        "10": 3,
        "11": 3,
    }


# ==============================================================================
# test_HierarchicalConstrainedClustering_compute_predicted_clusters_travelling_by_iteration
# ==============================================================================
def test_HierarchicalConstrainedClustering_compute_predicted_clusters_travelling_by_iteration():
    """
    Test that the `compute_predicted_clusters` method of the `HierarchicalConstrainedClustering` clustering works by travelling by `"iteration"`.
    """

    # Define `vectors` and `constraints_manager`
    vectors = {
        "00": csr_matrix([1.00, 0.00, 0.00]),
        "01": csr_matrix([0.99, 0.00, 0.00]),
        "02": csr_matrix([0.97, 0.00, 0.00]),
        "03": csr_matrix([0.96, 0.00, 0.00]),
        "04": csr_matrix([0.94, 0.00, 0.00]),
        "05": csr_matrix([0.93, 0.00, 0.00]),
        "06": csr_matrix([0.80, 0.80, 0.00]),
        "07": csr_matrix([0.80, 0.81, 0.00]),
        "08": csr_matrix([0.00, 0.00, 0.70]),
        "09": csr_matrix([0.00, 0.00, 0.71]),
        "10": csr_matrix([0.00, 0.00, 0.99]),
        "11": csr_matrix([0.00, 0.00, 1.00]),
    }
    constraints_manager = BinaryConstraintsManager(list_of_data_IDs=list(vectors.keys()))

    # Initialize a `HierarchicalConstrainedClustering` instance.
    clustering_model = HierarchicalConstrainedClustering(
        linkage="single",
        random_seed=1,
    )

    # Compute all clustering tree.
    clustering_model.cluster(
        constraints_manager=constraints_manager,
        vectors=vectors,
        nb_clusters=2,
    )

    # Run `compute_predicted_clusters` while travalleing clustering tree by `"iteration"`.
    dict_of_predicted_clusters = clustering_model.compute_predicted_clusters(
        nb_clusters=4,
        by="iteration",
    )
    assert dict_of_predicted_clusters == {
        "00": 0,
        "01": 0,
        "02": 0,
        "03": 0,
        "04": 0,
        "05": 0,
        "06": 1,
        "07": 1,
        "08": 2,
        "09": 2,
        "10": 3,
        "11": 3,
    }


# ==============================================================================
# test_HierarchicalConstrainedClustering_cluster_with_break_loop
# ==============================================================================
def test_HierarchicalConstrainedClustering_cluster_with_break_loop():
    """
    Test that the `clustering.hierarchical.HierarchicalConstrainedClustering` clustering can break clustering loop.
    """

    # Define `vectors` and `constraints_manager`
    vectors = {
        "0": csr_matrix([1.00, 0.00, 0.00, 0.00]),
        "1": csr_matrix([0.95, 0.02, 0.02, 0.01]),
        "2": csr_matrix([0.98, 0.00, 0.02, 0.00]),
        "3": csr_matrix([0.99, 0.00, 0.01, 0.00]),
        "4": csr_matrix([0.50, 0.22, 0.21, 0.07]),
        "5": csr_matrix([0.50, 0.21, 0.22, 0.07]),
        "6": csr_matrix([0.01, 0.01, 0.01, 0.97]),
        "7": csr_matrix([0.00, 0.01, 0.00, 0.99]),
        "8": csr_matrix([0.00, 0.00, 0.00, 1.00]),
    }
    constraints_manager = BinaryConstraintsManager(list_of_data_IDs=["0", "1", "2", "3", "4", "5", "6", "7", "8"])
    constraints_manager.add_constraint(data_ID1="0", data_ID2="1", constraint_type="CANNOT_LINK")
    constraints_manager.add_constraint(data_ID1="0", data_ID2="2", constraint_type="CANNOT_LINK")
    constraints_manager.add_constraint(data_ID1="0", data_ID2="3", constraint_type="CANNOT_LINK")
    constraints_manager.add_constraint(data_ID1="0", data_ID2="4", constraint_type="CANNOT_LINK")
    constraints_manager.add_constraint(data_ID1="0", data_ID2="5", constraint_type="CANNOT_LINK")
    constraints_manager.add_constraint(data_ID1="0", data_ID2="6", constraint_type="CANNOT_LINK")
    constraints_manager.add_constraint(data_ID1="0", data_ID2="7", constraint_type="CANNOT_LINK")
    constraints_manager.add_constraint(data_ID1="0", data_ID2="8", constraint_type="CANNOT_LINK")
    constraints_manager.add_constraint(data_ID1="1", data_ID2="2", constraint_type="CANNOT_LINK")
    constraints_manager.add_constraint(data_ID1="1", data_ID2="3", constraint_type="CANNOT_LINK")
    constraints_manager.add_constraint(data_ID1="1", data_ID2="4", constraint_type="CANNOT_LINK")
    constraints_manager.add_constraint(data_ID1="1", data_ID2="5", constraint_type="CANNOT_LINK")
    constraints_manager.add_constraint(data_ID1="1", data_ID2="6", constraint_type="CANNOT_LINK")
    constraints_manager.add_constraint(data_ID1="1", data_ID2="7", constraint_type="CANNOT_LINK")
    constraints_manager.add_constraint(data_ID1="1", data_ID2="8", constraint_type="CANNOT_LINK")
    constraints_manager.add_constraint(data_ID1="2", data_ID2="3", constraint_type="CANNOT_LINK")
    constraints_manager.add_constraint(data_ID1="2", data_ID2="4", constraint_type="CANNOT_LINK")
    constraints_manager.add_constraint(data_ID1="2", data_ID2="5", constraint_type="CANNOT_LINK")
    constraints_manager.add_constraint(data_ID1="2", data_ID2="6", constraint_type="CANNOT_LINK")
    constraints_manager.add_constraint(data_ID1="2", data_ID2="7", constraint_type="CANNOT_LINK")
    constraints_manager.add_constraint(data_ID1="2", data_ID2="8", constraint_type="CANNOT_LINK")
    constraints_manager.add_constraint(data_ID1="3", data_ID2="4", constraint_type="CANNOT_LINK")
    constraints_manager.add_constraint(data_ID1="3", data_ID2="5", constraint_type="CANNOT_LINK")
    constraints_manager.add_constraint(data_ID1="3", data_ID2="6", constraint_type="CANNOT_LINK")
    constraints_manager.add_constraint(data_ID1="3", data_ID2="7", constraint_type="CANNOT_LINK")
    constraints_manager.add_constraint(data_ID1="3", data_ID2="8", constraint_type="CANNOT_LINK")
    constraints_manager.add_constraint(data_ID1="4", data_ID2="5", constraint_type="CANNOT_LINK")
    constraints_manager.add_constraint(data_ID1="4", data_ID2="6", constraint_type="CANNOT_LINK")
    constraints_manager.add_constraint(data_ID1="4", data_ID2="7", constraint_type="CANNOT_LINK")
    constraints_manager.add_constraint(data_ID1="4", data_ID2="8", constraint_type="CANNOT_LINK")
    constraints_manager.add_constraint(data_ID1="5", data_ID2="6", constraint_type="CANNOT_LINK")
    constraints_manager.add_constraint(data_ID1="5", data_ID2="7", constraint_type="CANNOT_LINK")
    constraints_manager.add_constraint(data_ID1="5", data_ID2="8", constraint_type="CANNOT_LINK")
    constraints_manager.add_constraint(data_ID1="6", data_ID2="7", constraint_type="CANNOT_LINK")
    constraints_manager.add_constraint(data_ID1="6", data_ID2="8", constraint_type="CANNOT_LINK")
    constraints_manager.add_constraint(data_ID1="7", data_ID2="8", constraint_type="CANNOT_LINK")

    # Initialize a `HierarchicalConstrainedClustering` instance.
    clustering_model = HierarchicalConstrainedClustering(
        linkage="average",
    )

    # Run clustering.
    clustering_model.cluster(
        constraints_manager=constraints_manager,
        vectors=vectors,
        nb_clusters=5,
    )
    assert clustering_model.dict_of_predicted_clusters
    assert clustering_model.dict_of_predicted_clusters == {
        "0": 0,
        "1": 1,
        "2": 2,
        "3": 3,
        "4": 4,
        "5": 5,
        "6": 6,
        "7": 7,
        "8": 8,
    }


# ==============================================================================
# test_HierarchicalConstrainedClustering_cluster_end_cases_with_too_many_clusters
# ==============================================================================
def test_HierarchicalConstrainedClustering_cluster_end_cases_with_too_many_clusters():
    """
    Test that the `clustering.hierarchical.HierarchicalConstrainedClustering` clustering works with too many clusters.
    """

    # Define `vectors` and `constraints_manager`
    vectors = {
        "0": csr_matrix([1.00, 0.00, 0.00, 0.00]),
        "1": csr_matrix([0.95, 0.02, 0.02, 0.01]),
        "2": csr_matrix([0.98, 0.00, 0.02, 0.00]),
        "3": csr_matrix([0.99, 0.00, 0.01, 0.00]),
        "4": csr_matrix([0.50, 0.22, 0.21, 0.07]),
        "5": csr_matrix([0.50, 0.21, 0.22, 0.07]),
        "6": csr_matrix([0.01, 0.01, 0.01, 0.97]),
        "7": csr_matrix([0.00, 0.01, 0.00, 0.99]),
        "8": csr_matrix([0.00, 0.00, 0.00, 1.00]),
    }
    constraints_manager = BinaryConstraintsManager(list_of_data_IDs=list(vectors.keys()))

    # Initialize a `HierarchicalConstrainedClustering` instance.
    clustering_model = HierarchicalConstrainedClustering(
        linkage="average",
    )

    # Run clustering.
    dict_of_predicted_clusters = clustering_model.cluster(
        constraints_manager=constraints_manager,
        vectors=vectors,
        nb_clusters=99,
    )
    assert clustering_model.dict_of_predicted_clusters
    assert dict_of_predicted_clusters == {
        "0": 0,
        "1": 1,
        "2": 2,
        "3": 3,
        "4": 4,
        "5": 5,
        "6": 6,
        "7": 7,
        "8": 8,
    }


# ==============================================================================
# test_Cluster_for_inconsistent_children_and_members
# ==============================================================================
def test_Cluster_for_inconsistent_children_and_members():
    """
    Test that the `clustering.hierarchical.Cluster` initialization raises an `ValueError` for inconsistent `children` and `members` parameters.
    """

    # Define `vectors`.
    vectors = {
        "0": csr_matrix([1.00, 0.00]),
        "1": csr_matrix([0.99, 0.01]),
        "2": csr_matrix([0.02, 0.98]),
        "3": csr_matrix([0.01, 0.99]),
        "4": csr_matrix([0.00, 1.00]),
    }

    # Check `ValueError` for both `children` and `members` unset.
    with pytest.raises(ValueError, match="by `children` setting or by `members` setting"):
        Cluster(vectors=vectors, cluster_ID=2, clustering_iteration=1, children=None, members=None)

    # Check `ValueError` for both `children` and `members` set.
    with pytest.raises(ValueError, match="by `children` setting or by `members` setting"):
        Cluster(
            vectors=vectors,
            cluster_ID=2,
            clustering_iteration=1,
            children=[
                Cluster(
                    vectors=vectors,
                    cluster_ID=0,
                    clustering_iteration=0,
                    members=["0", "1"],
                ),
                Cluster(
                    vectors=vectors,
                    cluster_ID=1,
                    clustering_iteration=0,
                    members=["2", "3", "4"],
                ),
            ],
            members=["5", "6", "7", "8", "9"],
        )


# ==============================================================================
# test_Cluster_to_dict
# ==============================================================================
def test_Cluster_add_new_children():
    """
    Test that the `clustering.hierarchical.Cluster.add_new_children` method of `Cluster` class works.
    """

    # Define `vectors`.
    vectors = {
        "0": csr_matrix([1.00, 0.00]),
        "1": csr_matrix([0.99, 0.01]),
        "2": csr_matrix([0.02, 0.98]),
        "3": csr_matrix([0.01, 0.99]),
        "4": csr_matrix([0.00, 1.00]),
    }

    # Create `clusters`.
    clusters = Cluster(
        vectors=vectors,
        cluster_ID=2,
        clustering_iteration=1,
        children=[
            Cluster(
                vectors=vectors,
                cluster_ID=0,
                clustering_iteration=0,
                members=["0", "1"],
            ),
        ],
    )

    assert clusters.members == ["0", "1"]
    assert clusters.clustering_iteration == 1
    assert clusters.get_cluster_size() == 2

    clusters.add_new_children(
        new_children=[
            Cluster(
                vectors=vectors,
                cluster_ID=1,
                clustering_iteration=0,
                members=["2", "3", "4"],
            ),
        ],
        new_clustering_iteration=2,
    )

    assert clusters.members == ["0", "1", "2", "3", "4"]
    assert clusters.clustering_iteration == 2
    assert clusters.get_cluster_size() == 5


# ==============================================================================
# test_Cluster_to_dict
# ==============================================================================
def test_Cluster_to_dict():
    """
    Test that the `clustering.hierarchical.Cluster.to_dict` method of `Cluster` class works.
    """

    # Define `vectors`.
    vectors = {
        "0": csr_matrix([1.00, 0.00]),
        "1": csr_matrix([0.99, 0.01]),
        "2": csr_matrix([0.02, 0.98]),
        "3": csr_matrix([0.01, 0.99]),
        "4": csr_matrix([0.00, 1.00]),
    }

    # Create `clusters`.
    clusters = Cluster(
        vectors=vectors,
        cluster_ID=2,
        clustering_iteration=1,
        children=[
            Cluster(
                vectors=vectors,
                cluster_ID=0,
                clustering_iteration=0,
                members=["0", "1"],
            ),
            Cluster(
                vectors=vectors,
                cluster_ID=1,
                clustering_iteration=0,
                members=["2", "3", "4"],
            ),
        ],
    )

    # Define expected dictionnary.
    dict_expected = {
        "cluster_ID": 2,
        "clustering_iteration": 1,
        "children": [
            {
                "cluster_ID": 0,
                "clustering_iteration": 0,
                "children": [],
                "cluster_inverse_depth": 0,
                "members": ["0", "1"],
            },
            {
                "cluster_ID": 1,
                "clustering_iteration": 0,
                "children": [],
                "cluster_inverse_depth": 0,
                "members": ["2", "3", "4"],
            },
        ],
        "cluster_inverse_depth": 1,
        "members": ["0", "1", "2", "3", "4"],
    }

    assert clusters.to_dict() == dict_expected



# -*- coding: utf-8 -*-

"""
* Name:         cognitivefactory.interactive_clustering.constraints.abstract
* Description:  The abstract class used to define constraints managing algorithms.
* Author:       Erwan Schild
* Created:      17/03/2021
* Licence:      CeCILL-C License v1.0 (https://cecill.info/licences.fr.html)
"""

# ==============================================================================
# IMPORTS :
# ==============================================================================

from abc import ABC, abstractmethod  # To define an abstract class.
from typing import List, Optional, Tuple  # To type Python code (mypy).


# ==============================================================================
# ABSTRACT CONSTRAINTS MANAGING
# ==============================================================================
class AbstractConstraintsManager(ABC):
    """
    Abstract class that is used to define constraints manager.
    The main inherited methods are about data IDs management, constraints management and constraints exploration.

    References:
        - Constraints in clustering: `Wagstaff, K. et C. Cardie (2000). Clustering with Instance-level Constraints. Proceedings of the Seventeenth International Conference on Machine Learning, 1103–1110.`
    """

    # ==============================================================================
    # ABSTRACT METHOD - DATA_ID MANAGEMENT
    # ==============================================================================
    @abstractmethod
    def add_data_ID(
        self,
        data_ID: str,
    ) -> bool:
        """
        (ABSTRACT METHOD)
        An abstract method that represents the main method used to add a new data ID to manage.

        Args:
            data_ID (str): The data ID to manage.

        Raises:
            ValueError: if `data_ID` is already managed.

        Returns:
            bool: `True` if the addition is done.
        """

    @abstractmethod
    def delete_data_ID(
        self,
        data_ID: str,
    ) -> bool:
        """
        (ABSTRACT METHOD)
        An abstract method that represents the main method used to delete a data ID to no longer manage.

        Args:
            data_ID (str): The data ID to no longer manage.

        Raises:
            ValueError: if `data_ID` is not managed.

        Returns:
            bool: `True` if the deletion is done.
        """

    @abstractmethod
    def get_list_of_managed_data_IDs(
        self,
    ) -> List[str]:
        """
        (ABSTRACT METHOD)
        An abstract method that represents the main method used to get the list of data IDs that are managed.

        Returns:
            List[str]: The list of data IDs that are managed.
        """

    # ==============================================================================
    # ABSTRACT METHOD - CONSTRAINTS MANAGEMENT
    # ==============================================================================
    @abstractmethod
    def add_constraint(
        self,
        data_ID1: str,
        data_ID2: str,
        constraint_type: str,
        constraint_value: float = 1.0,
    ) -> bool:
        """
        (ABSTRACT METHOD)
        An abstract method that represents the main method used to add a constraint between two data IDs.

        Args:
            data_ID1 (str): The first data ID that is concerned for this constraint addition.
            data_ID2 (str): The second data ID that is concerned for this constraint addition.
            constraint_type (str): The type of the constraint to add. The type have to be `"MUST_LINK"` or `"CANNOT_LINK"`.
            constraint_value (float, optional): The value of the constraint to add. The value have to be in range `[0.0, 1.0]`. Defaults to 1.0.

        Raises:
            ValueError: if `data_ID1`, `data_ID2`, `constraint_type` are not managed, or if a conflict is detected with constraints inference.

        Returns:
            bool: `True` if the addition is done, `False` is the constraint can't be added.
        """

    @abstractmethod
    def delete_constraint(
        self,
        data_ID1: str,
        data_ID2: str,
    ) -> bool:
        """
        (ABSTRACT METHOD)
        An abstract method that represents the main method used to delete the constraint between two data IDs.

        Args:
            data_ID1 (str): The first data ID that is concerned for this constraint deletion.
            data_ID2 (str): The second data ID that is concerned for this constraint deletion.

        Raises:
            ValueError: if `data_ID1` or `data_ID2` are not managed.

        Returns:
            bool: `True` if the deletion is done.
        """

    @abstractmethod
    def get_added_constraint(
        self,
        data_ID1: str,
        data_ID2: str,
    ) -> Optional[Tuple[str, float]]:
        """
        (ABSTRACT METHOD)
        An abstract method that represents the main method used to get the constraint added between the two data IDs.
        Do not take into account the constraints transitivity, just look at constraints that are explicitly added.

        Args:
            data_ID1 (str): The first data ID that is concerned for this constraint.
            data_ID2 (str): The second data ID that is concerned for this constraint.

        Raises:
            ValueError: if `data_ID1` or `data_ID2` are not managed.

        Returns:
            Optional[Tuple[str, float]]: `None` if no constraint, `(constraint_type, constraint_value)` otherwise.
        """

    # ==============================================================================
    # ABSTRACT METHOD - CONSTRAINTS EXPLORATION
    # ==============================================================================
    @abstractmethod
    def get_inferred_constraint(
        self,
        data_ID1: str,
        data_ID2: str,
        threshold: float = 1.0,
    ) -> Optional[str]:
        """
        (ABSTRACT METHOD)
        An abstract method that represents the main method used to check if the constraint inferred by transitivity between the two data IDs.
        The transitivity is taken into account, and the `threshold` parameter is used to evaluate the impact of constraints transitivity.

        Args:
            data_ID1 (str): The first data ID that is concerned for this constraint.
            data_ID2 (str): The second data ID that is concerned for this constraint.
            threshold (float, optional): The threshold used to evaluate the impact of constraints transitivity link. Defaults to `1.0`.

        Raises:
            ValueError: if `data_ID1`, `data_ID2` or `threshold` are not managed.

        Returns:
            Optional[str]: The type of the inferred constraint. The type can be `None`, `"MUST_LINK"` or `"CANNOT_LINK"`.
        """

    @abstractmethod
    def get_connected_components(
        self,
        threshold: float = 1.0,
    ) -> List[List[str]]:
        """
        (ABSTRACT METHOD)
        An abstract method that represents the main method used to get the possible lists of data IDs that are connected by a `"MUST_LINK"` constraints.
        Each list forms a component of the constraints transitivity graph, and it forms a partition of the managed data IDs.
        The transitivity is taken into account, and the `threshold` parameter is used to evaluate the impact of constraints transitivity.

        Args:
            threshold (float, optional): The threshold used to evaluate the impact of constraints transitivity link. Defaults to `1.0`.

        Raises:
            ValueError: if `threshold` is not managed.

        Returns:
            List[List[int]]: The list of lists of data IDs that represent a component of the constraints transitivity graph.
        """

    @abstractmethod
    def check_completude_of_constraints(
        self,
        threshold: float = 1.0,
    ) -> bool:
        """
        (ABSTRACT METHOD)
        An abstract method that represents the main method used to check if all possible constraints are known (not necessarily annotated because of the transitivity).
        The transitivity is taken into account, and the `threshold` parameter is used to evaluate the impact of constraints transitivity.

        Args:
            threshold (float, optional): The threshold used to evaluate the impact of constraints transitivity link. Defaults to `1.0`.

        Raises:
            ValueError: if `threshold` is not managed.

        Returns:
            bool: Return `True` if all constraints are known, `False` otherwise.
        """

    @abstractmethod
    def get_min_and_max_number_of_clusters(
        self,
        threshold: float = 1.0,
    ) -> Tuple[int, int]:
        """
        (ABSTRACT METHOD)
        An abstract method that represents the main method used to get determine, for a clustering model that would not violate any constraints, the range of the possible clusters number.
        The transitivity is taken into account, and the `threshold` parameter is used to evaluate the impact of constraints transitivity.

        Args:
            threshold (float, optional): The threshold used to evaluate the impact of constraints transitivity link. Defaults to `1.0`.

        Raises:
            ValueError: if `threshold` is not managed.

        Returns:
            Tuple[int,int]: The minimum and the maximum possible clusters numbers (for a clustering model that would not violate any constraints).
        """

    # ==============================================================================
    # ABSTRACT METHOD - CONSTRAINTS CONFLICT
    # ==============================================================================

    @abstractmethod
    def get_list_of_involved_data_IDs_in_a_constraint_conflict(
        self,
        data_ID1: str,
        data_ID2: str,
        constraint_type: str,
    ) -> Optional[List[str]]:
        """
        (ABSTRACT METHOD)
        An abstract method that represents the main method used to get all data IDs involved in a constraints conflict.

        Args:
            data_ID1 (str): The first data ID involved in the constraint_conflit.
            data_ID2 (str): The second data ID involved in the constraint_conflit.
            constraint_type (str): The constraint that create a conflict. The constraints can be `"MUST_LINK"` or `"CANNOT_LINK"`.

        Raises:
            ValueError: if `data_ID1`, `data_ID2`, `constraint_type` are not managed.

        Returns:
            Optional[List[str]]: The list of data IDs that are involved in the conflict. It matches data IDs from connected components of `data_ID1` and `data_ID2`.
        """



# -*- coding: utf-8 -*-

"""
* Name:         tests/
* Description:  Tests suite for `cognitivefactory.interactive_clustering`.
* Author:       Erwan Schild
* Created:      17/03/2021
* Licence:      CeCILL (https://cecill.info/licences.fr.html)
"""

from pathlib import Path

TESTS_DIR = Path(__file__).parent
TMP_DIR = TESTS_DIR / "tmp"
FIXTURES_DIR = TESTS_DIR / "fixtures"



# -*- coding: utf-8 -*-

"""
* Name:         interactive-clustering/tests/clustering/test_abstract.py
* Description:  Unittests for the `clustering.abstract` module.
* Author:       Erwan Schild
* Created:      17/03/2021
* Licence:      CeCILL (https://cecill.info/licences.fr.html)
"""

# ==============================================================================
# IMPORT PYTHON DEPENDENCIES
# ==============================================================================

import pytest

from cognitivefactory.interactive_clustering.clustering.abstract import (
    AbstractConstrainedClustering,
    rename_clusters_by_order,
)


# ==============================================================================
# test_AbstractConstrainedClustering_is_abstract
# ==============================================================================
def test_AbstractConstrainedClustering_is_abstract():
    """
    Test that the `clustering.abstract.AbstractConstrainedClustering` class is abstract.
    """

    # Check `TypeError` for initialization of `AbstractConstrainedClustering`.
    with pytest.raises(TypeError, match="Can't instantiate abstract class"):
        AbstractConstrainedClustering()


# ==============================================================================
# test_rename_clusters_by_order
# ==============================================================================
def test_rename_clusters_by_order():
    """
    Test that the `clustering.abstract.test_rename_clusters_by_order` mathod works.
    """

    assert rename_clusters_by_order(clusters={"d": 2, "b": 0, "a": 1, "e": 0, "c": 1}) == {
        "a": 0,
        "b": 1,
        "c": 0,
        "d": 2,
        "e": 1,
    }



# -*- coding: utf-8 -*-

"""
* Name:         cognitivefactory.interactive_clustering.utils.frequency
* Description:  Utilities methods for frequency analysis.
* Author:       Erwan Schild
* Created:      17/03/2021
* Licence:      CeCILL (https://cecill.info/licences.fr.html)
"""

# ==============================================================================
# IMPORT PYTHON DEPENDENCIES
# ==============================================================================

from typing import Dict, List  # To type Python code (mypy).


# ==============================================================================
# COMPUTE FREQUENCY OF CLUSTERS :
# ==============================================================================
def compute_clusters_frequency(clustering_result: Dict[str, int]) -> Dict[int, float]:
    """
    Get the frequency of each cluster present in a clustering result.

    Args:
        clustering_result (Dict[str,int]): The dictionary that contains the predicted cluster for each data ID.

    Returns:
        Dict[int,float] : Frequency fo each predicted intent.
    """

    # Get the total number of data IDs.
    nb_of_data_IDs = len(clustering_result.keys())

    # Default case : No data, so no cluster.
    if nb_of_data_IDs == 0:
        return {}

    # Get possible clusters IDs.
    list_of_possible_cluster_IDs: List[int] = sorted(
        {clustering_result[data_ID] for data_ID in clustering_result.keys()}
    )

    # Compute frequency of clusters in `clustering_result`.
    frequence_of_clusters: Dict[int, float] = {
        cluster_ID: len([data_ID for data_ID in clustering_result if clustering_result[data_ID] == cluster_ID])
        / nb_of_data_IDs
        for cluster_ID in list_of_possible_cluster_IDs
    }

    # Return the frequence of clusters.
    return frequence_of_clusters



# -*- coding: utf-8 -*-

"""
* Name:         interactive-clustering/tests/sampling/test_factory_random.py
* Description:  Unittests for the `sampling.cluster_based` module, `"random"` sampler.
* Author:       Erwan Schild
* Created:      17/03/2021
* Licence:      CeCILL (https://cecill.info/licences.fr.html)
"""

# ==============================================================================
# IMPORT PYTHON DEPENDENCIES
# ==============================================================================

import pytest

from cognitivefactory.interactive_clustering.constraints.binary import BinaryConstraintsManager
from cognitivefactory.interactive_clustering.sampling.clusters_based import ClustersBasedConstraintsSampling


# ==============================================================================
# test_factory_random_sampler_for_correct_settings
# ==============================================================================
def test_factory_random_sampler_for_correct_settings():
    """
    Test that the `random sampler` works for correct settings.
    """

    # Check a correct initialization.
    sampler = ClustersBasedConstraintsSampling(random_seed=1)

    assert sampler
    assert sampler.random_seed == 1


# ==============================================================================
# test_factory_random_sampler_sample_for_incorrect_constraints_manager
# ==============================================================================
def test_factory_random_sampler_sample_for_incorrect_constraints_manager():
    """
    Test that the `random sampler` sampling raises `ValueError` for incorrect `constraints_manager`.
    """

    # Initialize a `random sampler` instance.
    sampler = ClustersBasedConstraintsSampling(random_seed=1)

    # Check sample with incorrect `constraints_manager`.
    with pytest.raises(ValueError, match="`constraints_manager`"):
        sampler.sample(
            constraints_manager=None,
            nb_to_select=None,
        )


# ==============================================================================
# test_factory_random_sampler_sample_for_incorrect_nb_to_select
# ==============================================================================
def test_factory_random_sampler_sample_for_incorrect_nb_to_select():
    """
    Test that the `random sampler` sampling raises `ValueError` for incorrect `nb_to_select`.
    """

    # Initialize a `random sampler` instance.
    sampler = ClustersBasedConstraintsSampling(random_seed=1)

    # Check sample with incorrect `nb_to_select`.
    with pytest.raises(ValueError, match="`nb_to_select`"):
        sampler.sample(
            constraints_manager=BinaryConstraintsManager(
                list_of_data_IDs=[
                    "bonjour",
                    "salut",
                    "coucou",
                    "au revoir",
                    "a bientôt",
                ]
            ),
            nb_to_select=None,
        )

    # Check sample with incorrect `nb_to_select`
    with pytest.raises(ValueError, match="`nb_to_select`"):
        sampler.sample(
            constraints_manager=BinaryConstraintsManager(
                list_of_data_IDs=[
                    "bonjour",
                    "salut",
                    "coucou",
                    "au revoir",
                    "a bientôt",
                ],
            ),
            nb_to_select=-99,
        )


# ==============================================================================
# test_factory_random_sampler_sample_for_zero_nb_to_select
# ==============================================================================
def test_factory_random_sampler_sample_for_zero_nb_to_select():
    """
    Test that the `random sampler` sampling works for zero `nb_to_select`.
    """

    # Initialize a `random sampler` instance.
    sampler = ClustersBasedConstraintsSampling(random_seed=1)

    # Check sample with zero `nb_to_select`
    assert not sampler.sample(
        constraints_manager=BinaryConstraintsManager(
            list_of_data_IDs=[
                "bonjour",
                "salut",
                "coucou",
                "au revoir",
                "a bientôt",
            ],
        ),
        nb_to_select=0,
    )


# ==============================================================================
# test_factory_random_sampler_sample_for_empty_constraints_manager
# ==============================================================================
def test_factory_random_sampler_sample_for_empty_constraints_manager():
    """
    Test that the `random sampler` sampling works for empty `constraints_manager`.
    """

    # Initialize a `random sampler` instance.
    sampler = ClustersBasedConstraintsSampling(random_seed=1)

    # Check sample with empty `constraints_manager`
    assert (
        sampler.sample(
            constraints_manager=BinaryConstraintsManager(
                list_of_data_IDs=[
                    "bonjour",
                    "salut",
                    "coucou",
                    "au revoir",
                    "a bientôt",
                ],
            ),
            nb_to_select=3,
        )
        == [("coucou", "salut"), ("bonjour", "coucou"), ("au revoir", "salut")]
    )


# ==============================================================================
# test_factory_random_sampler_sample_for_correct_constraints_manager
# ==============================================================================
def test_factory_random_sampler_sample_for_correct_constraints_manager():
    """
    Test that the `random sampler` sampling works for correct `constraints_manager`.
    """

    # Initialize a `random sampler` instance.
    sampler = ClustersBasedConstraintsSampling(random_seed=1)

    # Initialize a `BinaryConstraintsManager` instance
    constraints_manager = BinaryConstraintsManager(
        list_of_data_IDs=[
            "bonjour",
            "salut",
            "coucou",
            "au revoir",
            "a bientôt",
        ]
    )
    constraints_manager.add_constraint(data_ID1="bonjour", data_ID2="salut", constraint_type="MUST_LINK")
    constraints_manager.add_constraint(data_ID1="au revoir", data_ID2="a bientôt", constraint_type="MUST_LINK")

    # Check sample with correct `constraints_manager`
    assert (
        sampler.sample(
            constraints_manager=constraints_manager,
            nb_to_select=3,
        )
        == [("au revoir", "bonjour"), ("au revoir", "coucou"), ("bonjour", "coucou")]
    )


# ==============================================================================
# test_factory_random_sampler_sample_for_full_annotated_constraints_manager
# ==============================================================================
def test_factory_random_sampler_sample_for_full_annotated_constraints_manager():
    """
    Test that the `random sampler` sampling works for full annotated `constraints_manager`.
    """

    # Initialize a `random sampler` instance.
    sampler = ClustersBasedConstraintsSampling(random_seed=1)

    # Initialize a `BinaryConstraintsManager` instance
    constraints_manager = BinaryConstraintsManager(
        list_of_data_IDs=[
            "bonjour",
            "salut",
            "coucou",
            "au revoir",
            "a bientôt",
        ]
    )
    constraints_manager.add_constraint(data_ID1="bonjour", data_ID2="salut", constraint_type="MUST_LINK")
    constraints_manager.add_constraint(data_ID1="bonjour", data_ID2="coucou", constraint_type="MUST_LINK")
    constraints_manager.add_constraint(data_ID1="bonjour", data_ID2="au revoir", constraint_type="CANNOT_LINK")
    constraints_manager.add_constraint(data_ID1="au revoir", data_ID2="a bientôt", constraint_type="MUST_LINK")

    # Check sample for full annotated `constraints_manager`
    assert not sampler.sample(
        constraints_manager=constraints_manager,
        nb_to_select=3,
    )



# -*- coding: utf-8 -*-

"""
* Name:         interactive-clustering/tests/tests_docs.py
* Description:  Unittests for the documentation.
* Author:       Erwan Schild
* Created:      17/03/2021
* Licence:      CeCILL (https://cecill.info/licences.fr.html)
"""


# ==============================================================================
# test_docs_usage :
# ==============================================================================
def test_docs_usage():
    """
    Test the `usage` documentation.
    """

    # Import dependencies.
    from cognitivefactory.interactive_clustering.clustering.factory import (  # noqa: C0415 (not top level import, it's fine)
        clustering_factory,
    )
    from cognitivefactory.interactive_clustering.constraints.factory import (  # noqa: C0415 (not top level import, it's fine)
        managing_factory,
    )
    from cognitivefactory.interactive_clustering.sampling.factory import (  # noqa: C0415 (not top level import, it's fine)
        sampling_factory,
    )
    from cognitivefactory.interactive_clustering.utils.preprocessing import (  # noqa: C0415 (not top level import, it's fine)
        preprocess,
    )
    from cognitivefactory.interactive_clustering.utils.vectorization import (  # noqa: C0415 (not top level import, it's fine)
        vectorize,
    )

    ###
    ### Initialization step (iteration `0`)
    ###
    # Define dictionary of texts.
    dict_of_texts = {
        "0": "This is my first question.",
        "1": "This is my second item.",
        "2": "This is my third query.",
        "3": "This is my fourth issue.",
        # ...
        "N": "This is my last request.",
    }

    # Preprocess data.
    dict_of_preprocess_texts = preprocess(
        dict_of_texts=dict_of_texts,
        spacy_language_model="fr_core_news_sm",
    )  # Apply simple preprocessing. Spacy language model has to be installed. Other parameters are available.

    # Vectorize data.
    dict_of_vectors = vectorize(
        dict_of_texts=dict_of_preprocess_texts,
        vectorizer_type="tfidf",
    )  # Apply TF-IDF vectorization. Other parameters are available.

    # Create an instance of binary constraints manager.
    constraints_manager = managing_factory(
        manager="binary",
        list_of_data_IDs=list(dict_of_texts.keys()),
    )
    assert constraints_manager

    # Create an instance of constrained COP-kmeans clustering.
    clustering_model = clustering_factory(
        algorithm="kmeans",
        random_seed=1,
    )  # Other clustering algorithms are available.
    assert clustering_model

    # Run clustering.
    clustering_result = clustering_model.cluster(
        constraints_manager=constraints_manager,
        vectors=dict_of_vectors,
        nb_clusters=2,
    )
    assert clustering_result

    ###
    ### Iteration step (iteration `N`)
    ###

    # Check if all constraints are already annotated.
    is_finish = constraints_manager.check_completude_of_constraints()

    # Print result
    if is_finish:  # pragma: no cover
        print("All possible constraints are annotated. No more iteration can be run.")
        # break

    # Create an instance of random sampler.
    sampler = sampling_factory(
        algorithm="random",
        random_seed=None,
    )  # Other algorithms are available.

    # Sample constraints to annotated.
    selection = sampler.sample(
        constraints_manager=constraints_manager,
        nb_to_select=3,
        # clustering_result=clustering_result,  # Results from iteration `N-1`.
        # vectors=dict_of_vectors,
    )
    assert len(selection) == 3
    assert selection

    # Annotate constraints (manual operation).
    ANNOTATIONS = ["MUST_LINK", "CANNOT_LINK", None]
    list_of_annotation = [
        (data_ID1, data_ID2, ANNOTATIONS[i]) for i, (data_ID1, data_ID2) in enumerate(selection)
    ]  # List of triplets with format `(data_ID1, data_ID2, annotation_type)` where `annotation_type` can be "MUST_LINK" or "CANNOT_LINK".

    for annotation in list_of_annotation:

        # Get the annotation
        data_ID1, data_ID2, constraint_type = annotation

        # Add constraints
        try:
            constraints_manager.add_constraint(data_ID1=data_ID1, data_ID2=data_ID2, constraint_type=constraint_type)
        except ValueError as err:
            print(
                err
            )  # An error can occur if parameters are incorrect or if annotation is incompatible with previous annotation.

    # Get min and max range of clusters based on constraints.
    min_n, max_n = constraints_manager.get_min_and_max_number_of_clusters()

    # Choose the number of cluster.
    nb_clusters = int((min_n + max_n) / 2)  # or manual selection.

    # Create an instance of constrained COP-kmeans clustering.
    clustering_model = clustering_factory(
        algorithm="kmeans",
        random_seed=1,
    )  # Other clustering algorithms are available.
    assert clustering_model

    # Run clustering.
    clustering_result = clustering_model.cluster(
        constraints_manager=constraints_manager,  # Annotation since iteration `0`.
        nb_clusters=nb_clusters,
        vectors=dict_of_vectors,
    )  # Clustering results are corrected since the previous iteration.
    assert clustering_result



# -*- coding: utf-8 -*-

"""
* Name:         cognitivefactory.interactive_clustering.constraints.factory
* Description:  The factory method used to easily initialize a constraints manager.
* Author:       Erwan Schild
* Created:      17/03/2021
* Licence:      CeCILL-C License v1.0 (https://cecill.info/licences.fr.html)
"""

# ==============================================================================
# IMPORTS :
# =============================================================================

from typing import List  # To type Python code (mypy).

from cognitivefactory.interactive_clustering.constraints.abstract import (  # To use abstract interface.
    AbstractConstraintsManager,
)
from cognitivefactory.interactive_clustering.constraints.binary import (  # To use binary constraints manager implementation.
    BinaryConstraintsManager,
)


# ==============================================================================
# MANAGING FACTORY :
# ==============================================================================
def managing_factory(list_of_data_IDs: List[str], manager: str = "binary", **kargs) -> AbstractConstraintsManager:
    """
    A factory to create a new instance of a constraints manager.

    Args:
        list_of_data_IDs (List[str]): The list of data IDs to manage.
        manager (str, optional): The identification of constraints manager to instantiate. Can be "binary". Defaults to `"binary"`.
        **kargs (dict): Other parameters that can be used in the instantiation.

    Raises:
        ValueError: if `manager` is not implemented.

    Returns:
        AbstractConstraintsManager : An instance of constraints manager.

    Examples:
        ```python
        # Import.
        from cognitivefactory.interactive_clustering.constraints.factory import managing_factory

        # Create an instance of binary constraints manager.
        constraints_manager = managing_factory(
            list_of_data_IDs=["0", "1", "2", "3", "4"],
            manager="binary",
        )
        ```
    """

    # Check that the requested algorithm is implemented.
    if manager != "binary":  # TODO use `not in {"binary"}`.
        raise ValueError("The `manager` '" + str(manager) + "' is not implemented.")

    # Case of Binary Constraints Manager
    ## if manager=="binary":

    return BinaryConstraintsManager(list_of_data_IDs=list_of_data_IDs, **kargs)



# -*- coding: utf-8 -*-

"""
* Name:         cognitivefactory.interactive_clustering.clustering.kmeans
* Description:  Implementation of constrained kmeans clustering algorithms.
* Author:       Erwan Schild
* Created:      17/03/2021
* Licence:      CeCILL-C License v1.0 (https://cecill.info/licences.fr.html)
"""

# ==============================================================================
# IMPORT PYTHON DEPENDENCIES
# ==============================================================================

import random  # To shuffle data and set random seed.
from typing import Dict, List, Optional, Union  # To type Python code (mypy).

import numpy as np  # To handle float.
from numpy import ndarray  # To handle matrix and vectors.
from scipy.sparse import csr_matrix  # To handle matrix and vectors.
from sklearn.metrics import pairwise_distances  # To compute distance.

from cognitivefactory.interactive_clustering.clustering.abstract import (  # To use abstract interface.; To sort clusters after computation.
    AbstractConstrainedClustering,
    rename_clusters_by_order,
)
from cognitivefactory.interactive_clustering.constraints.abstract import (  # To manage constraints.
    AbstractConstraintsManager,
)


# ==============================================================================
# KMEANS CONSTRAINED CLUSTERING
# ==============================================================================
class KMeansConstrainedClustering(AbstractConstrainedClustering):
    """
    This class implements the kmeans constrained clustering.
    It inherits from `AbstractConstrainedClustering`.

    References:
        - KMeans Clustering: `MacQueen, J. (1967). Some methods for classification and analysis of multivariate observations. Proceedings of the fifth Berkeley symposium on mathematical statistics and probability 1(14), 281–297.`
        - Constrained _'COP'_ KMeans Clustering: `Wagstaff, K., C. Cardie, S. Rogers, et S. Schroedl (2001). Constrained K-means Clustering with Background Knowledge. International Conference on Machine Learning`

    Examples:
        ```python
        # Import.
        from scipy.sparse import csr_matrix
        from cognitivefactory.interactive_clustering.constraints.binary import BinaryConstraintsManager
        from cognitivefactory.interactive_clustering.clustering.kmeans import KMeansConstrainedClustering

        # Create an instance of kmeans clustering.
        clustering_model = KMeansConstrainedClustering(
            model="COP",
            random_seed=2,
        )

        # Define vectors.
        # NB : use cognitivefactory.interactive_clustering.utils to preprocess and vectorize texts.
        vectors = {
            "0": csr_matrix([1.00, 0.00, 0.00, 0.00]),
            "1": csr_matrix([0.95, 0.02, 0.02, 0.01]),
            "2": csr_matrix([0.98, 0.00, 0.02, 0.00]),
            "3": csr_matrix([0.99, 0.00, 0.01, 0.00]),
            "4": csr_matrix([0.50, 0.22, 0.21, 0.07]),
            "5": csr_matrix([0.50, 0.21, 0.22, 0.07]),
            "6": csr_matrix([0.01, 0.01, 0.01, 0.97]),
            "7": csr_matrix([0.00, 0.01, 0.00, 0.99]),
            "8": csr_matrix([0.00, 0.00, 0.00, 1.00]),
        }

        # Define constraints manager.
        constraints_manager = BinaryConstraintsManager(list_of_data_IDs=list(vectors.keys()))
        constraints_manager.add_constraint(data_ID1="0", data_ID2="1", constraint_type="MUST_LINK")
        constraints_manager.add_constraint(data_ID1="0", data_ID2="7", constraint_type="MUST_LINK")
        constraints_manager.add_constraint(data_ID1="0", data_ID2="8", constraint_type="MUST_LINK")
        constraints_manager.add_constraint(data_ID1="4", data_ID2="5", constraint_type="MUST_LINK")
        constraints_manager.add_constraint(data_ID1="0", data_ID2="4", constraint_type="CANNOT_LINK")
        constraints_manager.add_constraint(data_ID1="2", data_ID2="4", constraint_type="CANNOT_LINK")

        # Run clustering.
        dict_of_predicted_clusters = clustering_model(
            constraints_manager=constraints_manager,
            vectors=vectors,
            nb_clusters=3,
        )

        # Print results.
        print("Expected results", ";", {"0": 0, "1": 0, "2": 1, "3": 1, "4": 2, "5": 2, "6": 0, "7": 0, "8": 0,})
        print("Computed results", ":", dict_of_predicted_clusters)
        ```
    """

    # ==============================================================================
    # INITIALIZATION
    # ==============================================================================
    def __init__(
        self,
        model: str = "COP",
        max_iteration: int = 150,
        tolerance: float = 1e-4,
        random_seed: Optional[int] = None,
        **kargs,
    ) -> None:
        """
        The constructor for KMeans Constrainted Clustering class.

        Args:
            model (str, optional): The kmeans clustering model to use. Available kmeans models are `"COP"`. Defaults to `"COP"`.
            max_iteration (int, optional): The maximum number of kmeans iteration for convergence. Defaults to `150`.
            tolerance (float, optional): The tolerance for convergence computation. Defaults to `1e-4`.
            random_seed (Optional[int]): The random seed to use to redo the same clustering. Defaults to `None`.
            **kargs (dict): Other parameters that can be used in the instantiation.

        Raises:
            ValueError: if some parameters are incorrectly set.
        """

        # Store `self.`model`.
        if model != "COP":  # TODO use `not in {"COP"}`.
            raise ValueError("The `model` '" + str(model) + "' is not implemented.")
        self.model: str = model

        # Store 'self.max_iteration`.
        if max_iteration < 1:
            raise ValueError("The `max_iteration` must be greater than or equal to 1.")
        self.max_iteration: int = max_iteration

        # Store `self.tolerance`.
        if tolerance < 0:
            raise ValueError("The `tolerance` must be greater than 0.0.")
        self.tolerance: float = tolerance

        # Store `self.random_seed`.
        self.random_seed: Optional[int] = random_seed

        # Store `self.kargs` for kmeans clustering.
        self.kargs = kargs

        # Initialize `self.dict_of_predicted_clusters`.
        self.dict_of_predicted_clusters: Optional[Dict[str, int]] = None

    # ==============================================================================
    # MAIN - CLUSTER DATA
    # ==============================================================================
    def cluster(
        self,
        constraints_manager: AbstractConstraintsManager,
        vectors: Dict[str, Union[ndarray, csr_matrix]],
        nb_clusters: int,
        verbose: bool = False,
        **kargs,
    ) -> Dict[str, int]:
        """
        The main method used to cluster data with the KMeans model.

        Args:
            constraints_manager (AbstractConstraintsManager): A constraints manager over data IDs that will force clustering to respect some conditions during computation.
            vectors (Dict[str,Union[ndarray,csr_matrix]]): The representation of data vectors. The keys of the dictionary represents the data IDs. This keys have to refer to the list of data IDs managed by the `constraints_manager`. The value of the dictionary represent the vector of each data. Vectors can be dense (`numpy.ndarray`) or sparse (`scipy.sparse.csr_matrix`).
            nb_clusters (int): The number of clusters to compute.  #TODO Set defaults to None with elbow method or other method ?
            verbose (bool, optional): Enable verbose output. Defaults to `False`.
            **kargs (dict): Other parameters that can be used in the clustering.

        Raises:
            ValueError: if `vectors` and `constraints_manager` are incompatible, or if some parameters are incorrectly set.

        Returns:
            Dict[str,int]: A dictionary that contains the predicted cluster for each data ID.
        """

        ###
        ### GET PARAMETERS
        ###

        # Store `self.constraints_manager` and `self.list_of_data_IDs`.
        if not isinstance(constraints_manager, AbstractConstraintsManager):
            raise ValueError("The `constraints_manager` parameter has to be a `AbstractConstraintsManager` type.")
        self.constraints_manager: AbstractConstraintsManager = constraints_manager
        self.list_of_data_IDs: List[str] = self.constraints_manager.get_list_of_managed_data_IDs()

        # Store `self.vectors`.
        if not isinstance(vectors, dict):
            raise ValueError("The `vectors` parameter has to be a `dict` type.")
        self.vectors: Dict[str, Union[ndarray, csr_matrix]] = vectors

        # Store `self.nb_clusters`.
        if nb_clusters < 2:
            raise ValueError("The `nb_clusters` '" + str(nb_clusters) + "' must be greater than or equal to 2.")
        self.nb_clusters: int = nb_clusters

        ###
        ### RUN KMEANS CONSTRAINED CLUSTERING
        ###

        # Initialize `self.dict_of_predicted_clusters`.
        self.dict_of_predicted_clusters = None

        # Case of `"COP"` KMeans clustering.
        ##### DEFAULTS : if self.model=="COP":
        self.dict_of_predicted_clusters = self._clustering_kmeans_model_COP(verbose=verbose)

        ###
        ### RETURN PREDICTED CLUSTERS
        ###

        return self.dict_of_predicted_clusters

    # ==============================================================================
    # IMPLEMENTATION - COP KMEANS CLUSTERING
    # ==============================================================================
    def _clustering_kmeans_model_COP(self, verbose: bool = False) -> Dict[str, int]:
        """
        Implementation of COP-kmeans algorithm, based on constraints violation check during cluster affectation.

        References:
            - Constrained _'COP'_ KMeans Clustering: `Wagstaff, K., C. Cardie, S. Rogers, et S. Schroedl (2001). Constrained K-means Clustering with Background Knowledge. International Conference on Machine Learning`

        Args:
            verbose (bool, optional): Enable verbose output. Defaults is `False`.

        Returns:
            Dict[str,int]: A dictionary that contains the predicted cluster for each data ID.
        """

        ###
        ### INITIALIZATION OF CENTROIDS
        ###

        # Initialize `self.centroids`.
        self.centroids: Dict[int, Union[ndarray, csr_matrix]] = self.initialize_centroids()

        # Initialize clusters
        self.clusters: Dict[str, int] = {data_ID: -1 for data_ID in self.list_of_data_IDs}

        ###
        ### RUN KMEANS UNTIL CONVERGENCE
        ###

        # Define variable of convergence checks.
        converged: bool = False
        shift: float = float("Inf")
        current_iteration: int = 0

        # While no convergence
        while (not converged) and (current_iteration < self.max_iteration):

            # Increase current iteration number.
            current_iteration += 1

            ###
            ### ASSIGN DATA TO THE CLOSEST CLUSTER WITH CONSTRAINTS RESPECT
            ###

            # Initialize temporary results.
            new_clusters: Dict[str, int] = {data_ID: -1 for data_ID in self.list_of_data_IDs}

            # For each data ID.
            list_of_data_IDs_to_assign: List[str] = self.list_of_data_IDs.copy()

            # While all data aren't assigned.
            while list_of_data_IDs_to_assign:

                # Get a data_ID to assign
                data_ID_to_assign: str = list_of_data_IDs_to_assign.pop()

                # Get the list of not compatible cluster IDs for assignation
                not_compatible_cluster_IDs: List[int] = [
                    new_clusters[data_ID]
                    for data_ID in self.list_of_data_IDs
                    if (new_clusters[data_ID] != -1)
                    and (
                        self.constraints_manager.get_inferred_constraint(
                            data_ID1=data_ID_to_assign,
                            data_ID2=data_ID,
                        )
                        == "CANNOT_LINK"
                    )
                ]

                # Get the list of possible cluster IDs for assignation.
                possible_cluster_IDs: List[int] = [
                    cluster_ID for cluster_ID in self.centroids.keys() if cluster_ID not in not_compatible_cluster_IDs
                ]

                # If there is no possible cluster...
                if possible_cluster_IDs == []:  # noqa: WPS520

                    # Assign the data ID to a new cluster
                    new_clusters[data_ID_to_assign] = max(
                        max([cluster_ID for _, cluster_ID in new_clusters.items()]) + 1, self.nb_clusters
                    )

                # If there is possible clusters...
                else:

                    # Compute distance between data ID and all possible centroids.
                    distances_to_cluster_ID: Dict[float, int] = {
                        pairwise_distances(
                            X=self.vectors[data_ID_to_assign],
                            Y=self.centroids[cluster_ID],
                            metric="euclidean",  # TODO get different pairwise_distances config in **kargs
                        )[0][0].astype(np.float64): cluster_ID
                        for cluster_ID in possible_cluster_IDs
                    }

                    # Get the clostest cluster to data ID by distance.
                    min_distance: float = min(distances_to_cluster_ID)
                    new_clusters[data_ID_to_assign] = distances_to_cluster_ID[min_distance]

                # Assign all data ID that are linked by a `"MUST_LINK"` constraint to the new cluster.
                for data_ID in self.list_of_data_IDs:
                    if (
                        self.constraints_manager.get_inferred_constraint(
                            data_ID1=data_ID_to_assign,
                            data_ID2=data_ID,
                        )
                        == "MUST_LINK"
                    ):
                        if data_ID in list_of_data_IDs_to_assign:
                            list_of_data_IDs_to_assign.remove(data_ID)
                        new_clusters[data_ID] = new_clusters[data_ID_to_assign]

            ###
            ### COMPUTE NEW CENTROIDS
            ###

            # Compute new centroids
            new_centroids: Dict[int, Union[ndarray, csr_matrix]] = self.compute_centroids(clusters=new_clusters)

            ###
            ### CHECK CONVERGENCE
            ###

            # Check by centroids difference (with tolerance).
            if set(self.clusters.values()) == set(new_clusters.values()):

                # Compute shift between kmeans iterations.
                shift = sum(
                    pairwise_distances(X=self.centroids[cluster_ID], Y=new_centroids[cluster_ID], metric="euclidean")[
                        0
                    ][0]
                    for cluster_ID in self.centroids
                )

                # If shift is under tolerance : convergence !
                converged = shift < self.tolerance

            # Check if number of clusters have changed.
            else:

                # Uncomparable shift.
                shift = float("Inf")

                # Don't converge.
                converged = False

            ###
            ### APPLY NEW COMPUTATIONS
            ###

            # Affect new clusters.
            self.clusters = new_clusters

            # Affect new centroids.
            self.centroids = new_centroids

            # Verbose.
            if verbose and (current_iteration % 5 == 0):  # pragma: no cover
                # Verbose - Print progression status.
                print("    CLUSTERING_ITERATION=" + str(current_iteration), ",", "shift=" + str(shift))

        # Verbose.
        if verbose:  # pragma: no cover
            # Verbose - Print progression status.
            print("    CLUSTERING_ITERATION=" + str(current_iteration), ",", "converged=" + str(converged))

        # Rename cluster IDs by order.
        self.clusters = rename_clusters_by_order(clusters=self.clusters)
        self.centroids = self.compute_centroids(clusters=new_clusters)

        return self.clusters

    # ==============================================================================
    # INITIALIZATION OF CLUSTERS
    # ==============================================================================
    def initialize_centroids(
        self,
    ) -> Dict[int, Union[ndarray, csr_matrix]]:
        """
        Initialize the centroid of each cluster by a vector.
        The choice is based on a random selection among data to cluster.

        Returns:
            Dict[int, Union[ndarray,csr_matrix]]: A dictionary which represent each cluster by a centroid.
        """

        # Get the list of possible indices.
        indices: List[str] = self.list_of_data_IDs.copy()

        # Shuffle the list of possible indices.
        random.seed(self.random_seed)
        random.shuffle(indices)

        # Subset indices.
        indices = indices[: self.nb_clusters]

        # Set initial centroids based on vectors.
        centroids: Dict[int, Union[ndarray, csr_matrix]] = {
            cluster_ID: self.vectors[indices[cluster_ID]] for cluster_ID in range(self.nb_clusters)
        }

        # Return centroids.
        return centroids

    # ==============================================================================
    # COMPUTE NEW CENTROIDS
    # ==============================================================================
    def compute_centroids(
        self,
        clusters: Dict[str, int],
    ) -> Dict[int, Union[ndarray, csr_matrix]]:
        """
        Compute the centroids of each cluster.

        Args:
            clusters (Dict[str,int]): Current clusters assignation.

        Returns:
            Dict[int, Union[ndarray,csr_matrix]]: A dictionary which represent each cluster by a centroid.
        """

        # Initialize centroids.
        centroids: Dict[int, Union[ndarray, csr_matrix]] = {}

        # For all possible cluster ID.
        for cluster_ID in set(clusters.values()):

            # Compute cluster members.
            members_of_cluster_ID = [
                vector for data_ID, vector in self.vectors.items() if clusters[data_ID] == cluster_ID
            ]

            # Compute centroid.
            centroid_for_cluster_ID: Union[ndarray, csr_matrix] = sum(members_of_cluster_ID) / len(
                members_of_cluster_ID
            )

            # Add centroids.
            centroids[cluster_ID] = centroid_for_cluster_ID

        # Return centroids.
        return centroids



# -*- coding: utf-8 -*-

"""
* Name:         tests/constraints
* Description:  Tests suite for `cognitivefactory.interactive_clustering.constraints`.
* Author:       Erwan Schild
* Created:      17/03/2021
* Licence:      CeCILL (https://cecill.info/licences.fr.html)
"""



# -*- coding: utf-8 -*-

"""
* Name:         interactive-clustering/tests/sampling/test_cluster_based.py
* Description:  Unittests for the `sampling.cluster_based` module.
* Author:       Erwan Schild
* Created:      04/10/2021
* Licence:      CeCILL (https://cecill.info/licences.fr.html)
"""

# ==============================================================================
# IMPORT PYTHON DEPENDENCIES
# ==============================================================================

import pytest

from cognitivefactory.interactive_clustering.sampling.clusters_based import ClustersBasedConstraintsSampling


# ==============================================================================
# test_cluster_based_for_incorrect_clusters_restriction_parameter
# ==============================================================================
def test_cluster_based_for_incorrect_clusters_restriction_parameter():
    """
    Test that the `sampling.cluster_based.ClustersBasedConstraintsSampling` initialization raises an `ValueError` for incorrect `clusters_restriction` parameter.
    """

    # Check `ValueError` for bad string value for `clusters_restriction`.
    with pytest.raises(ValueError, match="`clusters_restriction`"):
        ClustersBasedConstraintsSampling(
            clusters_restriction="unknown",
        )


# ==============================================================================
# test_cluster_based_for_incorrect_distance_restriction_parameter
# ==============================================================================
def test_cluster_based_for_incorrect_distance_restriction_parameter():
    """
    Test that the `sampling.cluster_based.ClustersBasedConstraintsSampling` initialization raises an `ValueError` for incorrect `distance_restriction` parameter.
    """

    # Check `ValueError` for bad string value for `distance_restriction`.
    with pytest.raises(ValueError, match="`distance_restriction`"):
        ClustersBasedConstraintsSampling(
            distance_restriction="unknown",
        )


# ==============================================================================
# test_cluster_based_for_incorrect_without_added_constraints_parameter
# ==============================================================================
def test_cluster_based_for_incorrect_without_added_constraints_parameter():
    """
    Test that the `sampling.cluster_based.ClustersBasedConstraintsSampling` initialization raises an `ValueError` for incorrect `without_added_constraints` parameter.
    """

    # Check `ValueError` for bad string value for `without_added_constraints`.
    with pytest.raises(ValueError, match="`without_added_constraints`"):
        ClustersBasedConstraintsSampling(
            without_added_constraints="unknown",
        )


# ==============================================================================
# test_cluster_based_for_incorrect_without_inferred_constraints_parameter
# ==============================================================================
def test_cluster_based_for_incorrect_without_inferred_constraints_parameter():
    """
    Test that the `sampling.cluster_based.ClustersBasedConstraintsSampling` initialization raises an `ValueError` for incorrect `without_inferred_constraints` parameter.
    """

    # Check `ValueError` for bad string value for `without_inferred_constraints`.
    with pytest.raises(ValueError, match="`without_inferred_constraints`"):
        ClustersBasedConstraintsSampling(
            without_inferred_constraints="unknown",
        )



# -*- coding: utf-8 -*-

"""
* Name:         cognitivefactory.interactive_clustering.sampling.clusters_based
* Description:  Implementation of constraints sampling based on clusters information.
* Author:       Erwan Schild
* Created:      04/10/2021
* Licence:      CeCILL (https://cecill.info/licences.fr.html)
"""

# ==============================================================================
# IMPORT PYTHON DEPENDENCIES
# ==============================================================================

import random  # To shuffle data and set random seed.
from typing import Dict, List, Optional, Tuple, Union  # To type Python code (mypy).

from numpy import ndarray  # To handle matrix and vectors.
from scipy.sparse import csr_matrix, vstack  # To handle matrix and vectors.
from sklearn.metrics import pairwise_distances  # To compute distance.

from cognitivefactory.interactive_clustering.constraints.abstract import (  # To manage constraints.
    AbstractConstraintsManager,
)
from cognitivefactory.interactive_clustering.sampling.abstract import (  # To use abstract interface.
    AbstractConstraintsSampling,
)


# ==============================================================================
# CLUSTERS BASED CONSTRAINTS SAMPLING
# ==============================================================================
class ClustersBasedConstraintsSampling(AbstractConstraintsSampling):
    """
    This class implements the sampling of data IDs based on clusters information in order to annotate constraints.
    It inherits from `AbstractConstraintsSampling`.

    Examples:
        ```python
        # Import.
        from r_wnlp.interactive_clustering.constraints.binary import BinaryConstraintsManager
        from r_wnlp.interactive_clustering.sampling.clusters_based import ClustersBasedConstraintsSampling

        # Create an instance of random sampling.
        sampler = ClustersBasedConstraintsSampling(random_seed=1)

        # Define list of data IDs.
        list_of_data_IDs = ["bonjour", "salut", "coucou", "au revoir", "a bientôt",]

        # Define constraints manager (set it to None for no constraints).
        constraints_manager = BinaryConstraintsManager(
            list_of_data_IDs=list_of_data_IDs,
        )
        constraints_manager.add_constraint(data_ID1="bonjour", data_ID2="salut", constraint_type="MUST_LINK")
        constraints_manager.add_constraint(data_ID1="au revoir", data_ID2="a bientôt", constraint_type="MUST_LINK")

        # Run sampling.
        selection = sampler.sample(
            constraints_manager=constraints_manager,
            nb_to_select=3,
        )

        # Print results.
        print("Expected results", ";", [("au revoir", "bonjour"), ("bonjour", "coucou"), ("a bientôt", "coucou"),])
        print("Computed results", ":", selection)
        ```
    """

    # ==============================================================================
    # INITIALIZATION
    # ==============================================================================
    def __init__(
        self,
        random_seed: Optional[int] = None,
        clusters_restriction: Optional[str] = None,
        distance_restriction: Optional[str] = None,
        without_added_constraints: bool = True,
        without_inferred_constraints: bool = True,
        **kargs,
    ) -> None:
        """
        The constructor for Clusters Based Constraints Sampling class.

        Args:
            random_seed (Optional[int]): The random seed to use to redo the same sampling. Defaults to `None`.
            clusters_restriction (Optional[str]): Restrict the sampling with a cluster constraints. Can impose data IDs to be in `"same_cluster"` or `"different_clusters"`. Defaults to `None`.  # TODO: `"specific_clusters"`
            distance_restriction (Optional[str]): Restrict the sampling with a distance constraints. Can impose data IDs to be `"closest_neighbors"` or `""farthest_neighbors""`. Defaults to `None`.
            without_added_constraints (bool): Option to not sample the already added constraints. Defaults to `True`.
            without_inferred_constraints (bool): Option to not sample the deduced constraints from already added one. Defaults to `True`.
            **kargs (dict): Other parameters that can be used in the instantiation.

        Raises:
            ValueError: if some parameters are incorrectly set.
        """

        # Store `self.random_seed`.
        self.random_seed: Optional[int] = random_seed

        # Store clusters restriction.
        if clusters_restriction not in {None, "same_cluster", "different_clusters"}:
            raise ValueError("The `clusters_restriction` '" + str(clusters_restriction) + "' is not implemented.")
        self.clusters_restriction: Optional[str] = clusters_restriction

        # Store distance restriction.
        if distance_restriction not in {None, "closest_neighbors", "farthest_neighbors"}:
            raise ValueError("The `distance_restriction` '" + str(distance_restriction) + "' is not implemented.")
        self.distance_restriction: Optional[str] = distance_restriction

        # Store constraints restrictions.
        if not isinstance(without_added_constraints, bool):
            raise ValueError("The `without_added_constraints` must be boolean")
        self.without_added_constraints: bool = without_added_constraints
        if not isinstance(without_inferred_constraints, bool):
            raise ValueError("The `without_inferred_constraints` must be boolean")
        self.without_inferred_constraints: bool = without_inferred_constraints

    # ==============================================================================
    # MAIN - SAMPLE
    # ==============================================================================
    def sample(
        self,
        constraints_manager: AbstractConstraintsManager,
        nb_to_select: int,
        clustering_result: Optional[Dict[str, int]] = None,
        vectors: Optional[Dict[str, Union[ndarray, csr_matrix]]] = None,
        **kargs,
    ) -> List[Tuple[str, str]]:
        """
        The main method used to sample pairs of data IDs for constraints annotation.

        Args:
            constraints_manager (AbstractConstraintsManager): A constraints manager over data IDs.
            nb_to_select (int): The number of pairs of data IDs to sample.
            clustering_result (Optional[Dict[str,int]], optional): A dictionary that represents the predicted cluster for each data ID. The keys of the dictionary represents the data IDs. If `None`, no clustering result are used during the sampling. Defaults to `None`.
            vectors (Optional[Dict[str,Union[ndarray,csr_matrix]]], optional): vectors (Dict[str,Union[ndarray,csr_matrix]]): The representation of data vectors. The keys of the dictionary represents the data IDs. This keys have to refer to the list of data IDs managed by the `constraints_manager`. The value of the dictionary represent the vector of each data. Vectors can be dense (`numpy.ndarray`) or sparse (`scipy.sparse.csr_matrix`). If `None`, no vectors are used during the sampling. Defaults to `None`
            **kargs (dict): Other parameters that can be used in the sampling.

        Raises:
            ValueError: if some parameters are incorrectly set or incompatible.

        Returns:
            List[Tuple[str,str]]: A list of couple of data IDs.
        """

        ###
        ### GET PARAMETERS
        ###

        # Check `constraints_manager`.
        if not isinstance(constraints_manager, AbstractConstraintsManager):
            raise ValueError("The `constraints_manager` parameter has to be a `AbstractConstraintsManager` type.")
        self.constraints_manager: AbstractConstraintsManager = constraints_manager

        # Check `nb_to_select`.
        if not isinstance(nb_to_select, int) or (nb_to_select < 0):
            raise ValueError("The `nb_to_select` '" + str(nb_to_select) + "' must be greater than or equal to 0.")
        elif nb_to_select == 0:
            return []

        # If `self.cluster_restriction` is set, check `clustering_result` parameters.
        if self.clusters_restriction is not None:
            if not isinstance(clustering_result, dict):
                raise ValueError("The `clustering_result` parameter has to be a `Dict[str, int]` type.")
            self.clustering_result: Dict[str, int] = clustering_result

        # If `self.distance_restriction` is set, check `vectors` parameters.
        if self.distance_restriction is not None:
            if not isinstance(vectors, dict):
                raise ValueError("The `vectors` parameter has to be a `Dict[str, Union[ndarray, csr_matrix]]` type.")
            self.vectors: Dict[str, Union[ndarray, csr_matrix]] = vectors

        ###
        ### DEFINE POSSIBLE PAIRS OF DATA IDS
        ###

        # Initialize possible pairs of data IDs
        list_of_possible_pairs_of_data_IDs: List[Tuple[str, str]] = []

        # Loop over pairs of data IDs.
        for data_ID1 in self.constraints_manager.get_list_of_managed_data_IDs():
            for data_ID2 in self.constraints_manager.get_list_of_managed_data_IDs():

                # Select ordered pairs.
                if data_ID1 >= data_ID2:
                    continue

                # Check clusters restriction.
                if (
                    self.clusters_restriction == "same_cluster"
                    and self.clustering_result[data_ID1] != self.clustering_result[data_ID2]
                ) or (
                    self.clusters_restriction == "different_clusters"
                    and self.clustering_result[data_ID1] == self.clustering_result[data_ID2]
                ):
                    continue

                # Check known constraints.
                if (
                    self.without_added_constraints is True
                    and self.constraints_manager.get_added_constraint(data_ID1=data_ID1, data_ID2=data_ID2) is not None
                ) or (
                    self.without_inferred_constraints is True
                    and self.constraints_manager.get_inferred_constraint(data_ID1=data_ID1, data_ID2=data_ID2)
                    is not None
                ):
                    continue

                # Add the pair of data IDs.
                list_of_possible_pairs_of_data_IDs.append((data_ID1, data_ID2))

        ###
        ### SAMPLING
        ###

        # Precompute pairwise distances.
        if self.distance_restriction is not None:

            # Compute pairwise distances.
            matrix_of_pairwise_distances: ndarray = pairwise_distances(
                X=vstack(vector for vector in self.vectors.values()),
                metric="euclidean",  # TODO get different pairwise_distances config in **kargs
            )

            # Format pairwise distances in a dictionary.
            self.dict_of_pairwise_distances: Dict[str, Dict[str, float]] = {
                vector_ID1: {
                    vector_ID2: float(matrix_of_pairwise_distances[i01, i02])
                    for i02, vector_ID2 in enumerate(self.vectors.keys())
                }
                for i01, vector_ID1 in enumerate(self.vectors.keys())
            }

        # Set random seed.
        random.seed(self.random_seed)

        # Case of closest neightbors selection.
        if self.distance_restriction == "closest_neighbors":
            return sorted(
                list_of_possible_pairs_of_data_IDs,
                key=lambda combination: self.dict_of_pairwise_distances[combination[0]][combination[1]],
            )[:nb_to_select]

        # Case of farthest neightbors selection.
        if self.distance_restriction == "farthest_neighbors":
            return sorted(
                list_of_possible_pairs_of_data_IDs,
                key=lambda combination: self.dict_of_pairwise_distances[combination[0]][combination[1]],
                reverse=True,
            )[:nb_to_select]

        # (default) Case of random selection.
        return random.sample(
            list_of_possible_pairs_of_data_IDs, k=min(nb_to_select, len(list_of_possible_pairs_of_data_IDs))
        )



# -*- coding: utf-8 -*-

"""
* Name:         interactive-clustering/tests/constraints/test_factory.py
* Description:  Unittests for the `constraints.factory` module.
* Author:       Erwan Schild
* Created:      17/03/2021
* Licence:      CeCILL (https://cecill.info/licences.fr.html)
"""

# ==============================================================================
# IMPORT PYTHON DEPENDENCIES
# ==============================================================================

import pytest

from cognitivefactory.interactive_clustering.constraints.binary import BinaryConstraintsManager
from cognitivefactory.interactive_clustering.constraints.factory import managing_factory


# ==============================================================================
# test_managing_factory_for_not_implemented_constraints_manager
# ==============================================================================
def test_managing_factory_for_not_implemented_constraints_manager():
    """
    Test that the `constraints.factory.managing_factory` method raises an `ValueError` for not implemented constraints manager.
    """

    # Check `ValueError` for bad string value for `manager`.
    with pytest.raises(ValueError, match="`manager`"):
        managing_factory(manager="unknown", list_of_data_IDs=["first", "second", "third"])


# ==============================================================================
# test_managing_factory_for_binary_constraints_manager
# ==============================================================================
def test_managing_factory_for_binary_constraints_manager():
    """
    Test that the `constraints.factory.managing_factory` can initialize an instance of `BinaryConstraintsManager`.
    """

    # Check `hierarchical` clustering.
    constraints_manager = managing_factory(algorithm="binary", list_of_data_IDs=["first", "second", "third"])
    assert isinstance(constraints_manager, BinaryConstraintsManager)



# -*- coding: utf-8 -*-

"""
* Name:         interactive-clustering/tests/sampling/test_factory_random_in_same_cluster.py
* Description:  Unittests for the `sampling.cluster_based` module, `"random_in_same_cluster"` sampler.
* Author:       Erwan Schild
* Created:      17/03/2021
* Licence:      CeCILL (https://cecill.info/licences.fr.html)
"""

# ==============================================================================
# IMPORT PYTHON DEPENDENCIES
# ==============================================================================

import pytest

from cognitivefactory.interactive_clustering.constraints.binary import BinaryConstraintsManager
from cognitivefactory.interactive_clustering.sampling.clusters_based import ClustersBasedConstraintsSampling


# ==============================================================================
# test_factory_random_in_same_cluster_sampler_for_correct_settings
# ==============================================================================
def test_factory_random_in_same_cluster_sampler_for_correct_settings():
    """
    Test that the `random_in_same_cluster sampler` works for correct settings.
    """

    # Check a correct initialization.
    sampler = ClustersBasedConstraintsSampling(
        clusters_restriction="same_cluster",
        random_seed=1,
    )

    assert sampler
    assert sampler.random_seed == 1


# ==============================================================================
# test_factory_random_in_same_cluster_sampler_sample_for_incorrect_constraints_manager
# ==============================================================================
def test_factory_random_in_same_cluster_sampler_sample_for_incorrect_constraints_manager():
    """
    Test that the `random_in_same_cluster sampler` sampling raises `ValueError` for incorrect `constraints_manager`.
    """

    # Initialize a `random_in_same sampler` instance.
    sampler = ClustersBasedConstraintsSampling(
        clusters_restriction="same_cluster",
        random_seed=1,
    )

    # Check sample with incorrect `constraints_manager`.
    with pytest.raises(ValueError, match="`constraints_manager`"):
        sampler.sample(
            constraints_manager=None,
            nb_to_select=None,
        )


# ==============================================================================
# test_factory_random_in_same_cluster_sampler_sample_for_incorrect_nb_to_select
# ==============================================================================
def test_factory_random_in_same_cluster_sampler_sample_for_incorrect_nb_to_select():
    """
    Test that the `random_in_same_cluster sampler` sampling raises `ValueError` for incorrect `nb_to_select`.
    """

    # Initialize a `random_in_same sampler` instance.
    sampler = ClustersBasedConstraintsSampling(
        clusters_restriction="same_cluster",
        random_seed=1,
    )

    # Check sample with incorrect `nb_to_select`.
    with pytest.raises(ValueError, match="`nb_to_select`"):
        sampler.sample(
            constraints_manager=BinaryConstraintsManager(
                list_of_data_IDs=[
                    "bonjour",
                    "salut",
                    "coucou",
                    "au revoir",
                    "a bientôt",
                ]
            ),
            nb_to_select=None,
        )

    # Check sample with incorrect `nb_to_select`
    with pytest.raises(ValueError, match="`nb_to_select`"):
        sampler.sample(
            constraints_manager=BinaryConstraintsManager(
                list_of_data_IDs=[
                    "bonjour",
                    "salut",
                    "coucou",
                    "au revoir",
                    "a bientôt",
                ],
            ),
            nb_to_select=-99,
        )


# ==============================================================================
# test_factory_random_in_same_cluster_sampler_sample_for_zero_nb_to_select
# ==============================================================================
def test_factory_random_in_same_cluster_sampler_sample_for_zero_nb_to_select():
    """
    Test that the `random_in_same_cluster sampler` sampling works for zero `nb_to_select`.
    """

    # Initialize a `random_in_same sampler` instance.
    sampler = ClustersBasedConstraintsSampling(
        clusters_restriction="same_cluster",
        random_seed=1,
    )

    # Check sample with zero `nb_to_select`.
    assert not sampler.sample(
        constraints_manager=BinaryConstraintsManager(
            list_of_data_IDs=[
                "bonjour",
                "salut",
                "coucou",
                "au revoir",
                "a bientôt",
            ],
        ),
        nb_to_select=0,
    )


# ==============================================================================
# test_factory_random_in_same_cluster_sampler_sample_for_incorrect_clustering_result
# ==============================================================================
def test_factory_random_in_same_cluster_sampler_sample_for_incorrect_clustering_result():
    """
    Test that the `random_in_same_cluster sampler` sampling raises `ValueError` or `KeyError` for incorrect `clustering_result`.
    """

    # Initialize a `random_in_same sampler` instance.
    sampler = ClustersBasedConstraintsSampling(
        clusters_restriction="same_cluster",
        random_seed=1,
    )

    # Check sample with incorrect `clustering_result`.
    with pytest.raises(ValueError, match="`clustering_result`"):
        sampler.sample(
            constraints_manager=BinaryConstraintsManager(
                list_of_data_IDs=[
                    "bonjour",
                    "salut",
                    "coucou",
                    "au revoir",
                    "a bientôt",
                ],
            ),
            nb_to_select=3,
            clustering_result="unknown",
        )

    # Check sample with incorrect `clustering_result`.
    with pytest.raises(KeyError, match="'a bientôt'|'au revoir'|'bonjour'|'coucou'|'salut'"):
        sampler.sample(
            constraints_manager=BinaryConstraintsManager(
                list_of_data_IDs=[
                    "bonjour",
                    "salut",
                    "coucou",
                    "au revoir",
                    "a bientôt",
                ],
            ),
            nb_to_select=3,
            clustering_result={
                "first": 1,
                "second": 2,
            },
        )


# ==============================================================================
# test_factory_random_in_same_cluster_sampler_sample_for_empty_constraints_manager
# ==============================================================================
def test_factory_random_in_same_cluster_sampler_sample_for_empty_constraints_manager():
    """
    Test that the `random_in_same_cluster sampler` sampling works for empty `constraints_manager`.
    """

    # Initialize a `random_in_same sampler` instance.
    sampler = ClustersBasedConstraintsSampling(
        clusters_restriction="same_cluster",
        random_seed=1,
    )

    # Check sample with empty `constraints_manager`.
    assert (
        sampler.sample(
            constraints_manager=BinaryConstraintsManager(
                list_of_data_IDs=[
                    "bonjour",
                    "salut",
                    "coucou",
                    "au revoir",
                    "a bientôt",
                ],
            ),
            nb_to_select=3,
            clustering_result={
                "bonjour": 0,
                "salut": 0,
                "coucou": 0,
                "au revoir": 1,
                "a bientôt": 1,
            },
        )
        == [("bonjour", "coucou"), ("coucou", "salut"), ("bonjour", "salut")]
    )


# ==============================================================================
# test_factory_random_in_same_cluster_sampler_sample_for_correct_constraints_manager
# ==============================================================================
def test_factory_random_in_same_cluster_sampler_sample_for_correct_constraints_manager():
    """
    Test that the `random_in_same_cluster sampler` sampling works for correct `constraints_manager`.
    """

    # Initialize a `random_in_same sampler` instance.
    sampler = ClustersBasedConstraintsSampling(
        clusters_restriction="same_cluster",
        random_seed=1,
    )

    # Initialize a `BinaryConstraintsManager` instance.
    constraints_manager = BinaryConstraintsManager(
        list_of_data_IDs=[
            "bonjour",
            "salut",
            "coucou",
            "au revoir",
            "a bientôt",
        ]
    )
    constraints_manager.add_constraint(data_ID1="bonjour", data_ID2="salut", constraint_type="MUST_LINK")
    constraints_manager.add_constraint(data_ID1="au revoir", data_ID2="a bientôt", constraint_type="MUST_LINK")

    # Check sample with correct `constraints_manager`.
    assert (
        sampler.sample(
            constraints_manager=constraints_manager,
            nb_to_select=3,
            clustering_result={
                "bonjour": 0,
                "salut": 0,
                "coucou": 0,
                "au revoir": 1,
                "a bientôt": 1,
            },
        )
        == [("bonjour", "coucou"), ("coucou", "salut")]
    )


# ==============================================================================
# test_factory_random_in_same_cluster_sampler_sample_for_full_annotated_constraints_manager
# ==============================================================================
def test_factory_random_in_same_cluster_sampler_sample_for_full_annotated_constraints_manager():
    """
    Test that the `random_in_same_cluster sampler` sampling works for full annotated `constraints_manager`.
    """

    # Initialize a `random_in_same sampler` instance.
    sampler = ClustersBasedConstraintsSampling(
        clusters_restriction="same_cluster",
        random_seed=1,
    )

    # Initialize a `BinaryConstraintsManager` instance.
    constraints_manager = BinaryConstraintsManager(
        list_of_data_IDs=[
            "bonjour",
            "salut",
            "coucou",
            "au revoir",
            "a bientôt",
        ]
    )
    constraints_manager.add_constraint(data_ID1="bonjour", data_ID2="salut", constraint_type="MUST_LINK")
    constraints_manager.add_constraint(data_ID1="bonjour", data_ID2="coucou", constraint_type="MUST_LINK")
    constraints_manager.add_constraint(data_ID1="bonjour", data_ID2="au revoir", constraint_type="CANNOT_LINK")
    constraints_manager.add_constraint(data_ID1="au revoir", data_ID2="a bientôt", constraint_type="MUST_LINK")

    # Check sample for full annotated `constraints_manager`.
    assert not sampler.sample(
        constraints_manager=constraints_manager,
        nb_to_select=3,
        clustering_result={
            "bonjour": 0,
            "salut": 0,
            "coucou": 0,
            "au revoir": 1,
            "a bientôt": 1,
        },
    )



# -*- coding: utf-8 -*-

"""
* Name:         cognitivefactory.interactive_clustering.clustering.hierarchical
* Description:  Implementation of constrained hierarchical clustering algorithms.
* Author:       Erwan Schild
* Created:      17/03/2021
* Licence:      CeCILL-C License v1.0 (https://cecill.info/licences.fr.html)
"""

# ==============================================================================
# IMPORT PYTHON DEPENDENCIES
# ==============================================================================

from datetime import datetime  # To use date in verbose output.
from typing import Any, Dict, List, Optional, Tuple, Union  # To type Python code (mypy).

import numpy as np  # To handle float.
from numpy import ndarray  # To handle matrix and vectors.
from scipy.sparse import csr_matrix  # To handle matrix and vectors.
from sklearn.metrics import pairwise_distances  # To compute distance.

from cognitivefactory.interactive_clustering.clustering.abstract import (  # To use abstract interface.; To sort clusters after computation.
    AbstractConstrainedClustering,
    rename_clusters_by_order,
)
from cognitivefactory.interactive_clustering.constraints.abstract import (  # To manage constraints.
    AbstractConstraintsManager,
)


# ==============================================================================
# HIERARCHICAL CONSTRAINED CLUSTERING
# ==============================================================================
class HierarchicalConstrainedClustering(AbstractConstrainedClustering):
    """
    This class implements the hierarchical constrained clustering.
    It inherits from `AbstractConstrainedClustering`.

    References:
        - Hierarchical Clustering: `Murtagh, F. et P. Contreras (2012). Algorithms for hierarchical clustering : An overview. Wiley Interdisc. Rew.: Data Mining and Knowledge Discovery 2, 86–97.`
        - Constrained Hierarchical Clustering: `Davidson, I. et S. S. Ravi (2005). Agglomerative Hierarchical Clustering with Constraints : Theoretical and Empirical Results. Springer, Berlin, Heidelberg 3721, 12.`

    Examples:
        ```python
        # Import.
        from scipy.sparse import csr_matrix
        from cognitivefactory.interactive_clustering.clustering.hierarchical import HierarchicalConstrainedClustering

        # Create an instance of hierarchical clustering.
        clustering_model = HierarchicalConstrainedClustering(
            linkage="ward",
            random_seed=2,
        )

        # Define vectors.
        # NB : use cognitivefactory.interactive_clustering.utils to preprocess and vectorize texts.
        vectors = {
            "0": csr_matrix([1.00, 0.00, 0.00]),
            "1": csr_matrix([0.95, 0.02, 0.01]),
            "2": csr_matrix([0.98, 0.00, 0.00]),
            "3": csr_matrix([0.99, 0.00, 0.00]),
            "4": csr_matrix([0.01, 0.99, 0.07]),
            "5": csr_matrix([0.02, 0.99, 0.07]),
            "6": csr_matrix([0.01, 0.99, 0.02]),
            "7": csr_matrix([0.01, 0.01, 0.97]),
            "8": csr_matrix([0.00, 0.01, 0.99]),
            "9": csr_matrix([0.00, 0.00, 1.00]),
        }

        # Define constraints manager.
        constraints_manager = BinaryConstraintsManager(list_of_data_IDs=list(vectors.keys()))

        # Run clustering.
        dict_of_predicted_clusters = clustering_model(
            constraints_manager=constraints_manager,
            vectors=vectors,
            nb_clusters=3,
        )

        # Print results.
        print("Expected results", ";", {"0": 0, "1": 0, "2": 0, "3": 0, "4": 1, "5": 1, "6": 1, "7": 2, "8": 2, "9": 2,})
        print("Computed results", ":", dict_of_predicted_clusters)
        ```
    """

    # ==============================================================================
    # INITIALIZATION
    # ==============================================================================
    def __init__(self, linkage: str = "ward", random_seed: Optional[int] = None, **kargs) -> None:
        """
        The constructor for Hierarchical Constrainted Clustering class.

        Args:
            linkage (str, optional): The metric used to merge clusters. Several type are implemented :
                - `"ward"`: Merge the two clusters for which the merged cluster from these clusters have the lowest intra-class distance.
                - `"average"`: Merge the two clusters that have the closest barycenters.
                - `"complete"`: Merge the two clusters for which the maximum distance between two data of these clusters is the lowest.
                - `"single"`: Merge the two clusters for which the minimum distance between two data of these clusters is the lowest.
                Defaults to `"ward"`.
            random_seed (Optional[int], optional): The random seed to use to redo the same clustering. Defaults to `None`.
            **kargs (dict): Other parameters that can be used in the instantiation.

        Raises:
            ValueError: if some parameters are incorrectly set.
        """

        # Store `self.linkage`.
        if linkage not in {"ward", "average", "complete", "single"}:
            raise ValueError("The `linkage` '" + str(linkage) + "' is not implemented.")
        self.linkage: str = linkage

        # Store `self.random_seed`
        self.random_seed: Optional[int] = random_seed

        # Store `self.kargs` for hierarchical clustering.
        self.kargs = kargs

        # Initialize `self.clustering_root` and `self.dict_of_predicted_clusters`.
        self.clustering_root: Optional[Cluster] = None
        self.dict_of_predicted_clusters: Optional[Dict[str, int]] = None

    # ==============================================================================
    # MAIN - CLUSTER DATA
    # ==============================================================================
    def cluster(
        self,
        constraints_manager: AbstractConstraintsManager,
        vectors: Dict[str, Union[ndarray, csr_matrix]],
        nb_clusters: int,
        verbose: bool = False,
        **kargs,
    ) -> Dict[str, int]:
        """
        The main method used to cluster data with the Hierarchical model.

        Args:
            constraints_manager (AbstractConstraintsManager): A constraints manager over data IDs that will force clustering to respect some conditions during computation.
            vectors (Dict[str,Union[ndarray,csr_matrix]]): The representation of data vectors. The keys of the dictionary represents the data IDs. This keys have to refer to the list of data IDs managed by the `constraints_manager`. The value of the dictionary represent the vector of each data. Vectors can be dense (`numpy.ndarray`) or sparse (`scipy.sparse.csr_matrix`).
            nb_clusters (int): The number of clusters to compute.  #TODO Set defaults to None with elbow method or other method ?
            verbose (bool, optional): Enable verbose output. Defaults to `False`.
            **kargs (dict): Other parameters that can be used in the clustering.

        Raises:
            ValueError: If some parameters are incorrectly set.

        Returns:
            Dict[str,int]: A dictionary that contains the predicted cluster for each data ID.
        """

        ###
        ### GET PARAMETERS
        ###

        # Store `self.constraints_manager` and `self.list_of_data_IDs`.
        if not isinstance(constraints_manager, AbstractConstraintsManager):
            raise ValueError("The `constraints_manager` parameter has to be a `AbstractConstraintsManager` type.")
        self.constraints_manager: AbstractConstraintsManager = constraints_manager
        self.list_of_data_IDs: List[str] = self.constraints_manager.get_list_of_managed_data_IDs()

        # Store `self.vectors`.
        if not isinstance(vectors, dict):
            raise ValueError("The `vectors` parameter has to be a `dict` type.")
        self.vectors: Dict[str, Union[ndarray, csr_matrix]] = vectors

        # Store `self.nb_clusters`.
        if nb_clusters < 2:
            raise ValueError("The `nb_clusters` '" + str(nb_clusters) + "' must be greater than or equal to 2.")
        self.nb_clusters: int = nb_clusters

        # Store `self.pairwise_distances_matrix`.
        self.pairwise_distances_matrix: Dict[str, Dict[str, float]] = {
            data_ID1: {
                data_ID2: pairwise_distances(X=self.vectors[data_ID1], Y=self.vectors[data_ID2], metric="euclidean",)[
                    0
                ][0].astype(np.float64)
                for data_ID2 in self.list_of_data_IDs
            }
            for data_ID1 in self.list_of_data_IDs
        }  # TODO need ?

        ###
        ### INITIALIZE HIERARCHICAL CONSTRAINED CLUSTERING
        ###

        # Verbose
        if verbose:  # pragma: no cover

            # Verbose - Print progression status.
            TIME_start: datetime = datetime.now()
            print(
                "    ",
                "CLUSTERING_ITERATION=" + "INITIALIZATION",
                "(current_time = " + str(TIME_start - TIME_start).split(".")[0] + ")",
            )

        # Initialize `self.clustering_root` and `self.dict_of_predicted_clusters`.
        self.clustering_root = None
        self.dict_of_predicted_clusters = None

        # Initialize iteration counter.
        self.clustering_iteration: int = 0

        # Initialize `current_clusters` and `self.clusters_storage`.
        self.current_clusters: List[int] = []
        self.clusters_storage: Dict[int, Cluster] = {}

        # Get the list of possibles lists of MUST_LINK data for initialization.
        list_of_possible_lists_of_MUST_LINK_data: List[List[str]] = self.constraints_manager.get_connected_components()

        # Estimation of max number of iteration.
        max_clustering_iteration: int = len(list_of_possible_lists_of_MUST_LINK_data) - 1

        # For each list of same data (MUST_LINK constraints).
        for MUST_LINK_data in list_of_possible_lists_of_MUST_LINK_data:

            # Create a initial cluster with data that MUST be LINKed.
            self._add_new_cluster_by_setting_members(
                members=MUST_LINK_data,
            )

        # Initialize distance between clusters.
        self.clusters_distance: Dict[int, Dict[int, float]] = {}
        for cluster_IDi in self.current_clusters:
            for cluster_IDj in self.current_clusters:
                if cluster_IDi < cluster_IDj:
                    # Compute distance between cluster i and cluster j.
                    distance: float = self._compute_distance(cluster_IDi=cluster_IDi, cluster_IDj=cluster_IDj)
                    # Store distance between cluster i and cluster j.
                    self._set_distance(cluster_IDi=cluster_IDi, cluster_IDj=cluster_IDj, distance=distance)

        # Initialize iterations at first iteration.
        self.clustering_iteration = 1

        ###
        ### RUN ITERATIONS OF HIERARCHICAL CONSTRAINED CLUSTERING UNTIL CONVERGENCE
        ###

        # Iter until convergence of clustering.
        while len(self.current_clusters) > 1:

            # Verbose
            if verbose:  # pragma: no cover

                # Verbose - Print progression status.
                TIME_current: datetime = datetime.now()
                print(
                    "    ",
                    "CLUSTERING_ITERATION="
                    + str(self.clustering_iteration).zfill(6)
                    + "/"
                    + str(max_clustering_iteration).zfill(6),
                    "(current_time = " + str(TIME_current - TIME_start).split(".")[0] + ")",
                    end="\r",
                )

            # Get clostest clusters to merge
            clostest_clusters: Optional[Tuple[int, int]] = self._get_the_two_clostest_clusters()

            # If no clusters to merge, then stop iterations.
            if clostest_clusters is None:
                break

            # Merge clusters the two closest clusters and add the merged cluster to the storage.
            # If merge one cluster "node" with a cluster "leaf" : add the cluster "leaf" to the children of the cluster "node".
            # If merge two clusters "nodes" or two clusters "leaves" : create a new cluster "node".
            merged_cluster_ID: int = self._add_new_cluster_by_merging_clusters(
                children=[
                    clostest_clusters[0],
                    clostest_clusters[1],
                ]
            )

            # Update distances
            for cluster_ID in self.current_clusters:

                if cluster_ID != merged_cluster_ID:
                    # Compute distance between cluster and merged cluster.
                    distance = self._compute_distance(cluster_IDi=cluster_ID, cluster_IDj=merged_cluster_ID)
                    # Store distance between cluster and merged cluster.
                    self._set_distance(cluster_IDi=cluster_ID, cluster_IDj=merged_cluster_ID, distance=distance)

            # Update self.clustering_iteration.
            self.clustering_iteration += 1

        ###
        ### END HIERARCHICAL CONSTRAINED CLUSTERING
        ###

        # Verbose
        if verbose:  # pragma: no cover

            # Verbose - Print progression status.
            TIME_current = datetime.now()

            # Case of clustering not completed.
            if len(self.current_clusters) > 1:
                print(
                    "    ",
                    "CLUSTERING_ITERATION=" + str(self.clustering_iteration).zfill(5),
                    "-",
                    "End : No more cluster to merge",
                    "(current_time = " + str(TIME_current - TIME_start).split(".")[0] + ")",
                )
            else:
                print(
                    "    ",
                    "CLUSTERING_ITERATION=" + str(self.clustering_iteration).zfill(5),
                    "-",
                    "End : Full clustering done",
                    "(current_time = " + str(TIME_current - TIME_start).split(".")[0] + ")",
                )

        # If several clusters remains, then merge them in a cluster root.
        if len(self.current_clusters) > 1:

            # Merge all remaining clusters.
            # If merge one cluster "node" with many cluster "leaves" : add clusters "leaves" to the children of the cluster "node".
            # If merge many clusters "nodes" and/or many clusters "leaves" : create a new cluster "node".
            self._add_new_cluster_by_merging_clusters(children=self.current_clusters.copy())

        # Get clustering root.
        root_ID: int = self.current_clusters[0]
        self.clustering_root = self.clusters_storage[root_ID]

        ###
        ### GET PREDICTED CLUSTERS
        ###

        # Compute predicted clusters.
        self.dict_of_predicted_clusters = self.compute_predicted_clusters(
            nb_clusters=self.nb_clusters,
        )

        return self.dict_of_predicted_clusters

    # ==============================================================================
    # ADD CLUSTER BY SETTING MEMBERS :
    # ==============================================================================
    def _add_new_cluster_by_setting_members(
        self,
        members: List[str],
    ) -> int:
        """
        Create or Update a cluster by setting its members, and add it to the storage and current clusters.

        Args:
            members (List[str]): A list of data IDs to define the new cluster by the data it contains.

        Returns:
            int : ID of the merged cluster.
        """
        # Get the ID of the new cluster.
        new_cluster_ID: int = max(self.clusters_storage.keys()) + 1 if (self.clusters_storage) else 0

        # Create the cluster.
        new_cluster = Cluster(
            vectors=self.vectors,
            cluster_ID=new_cluster_ID,
            clustering_iteration=self.clustering_iteration,
            members=members,
        )

        # Add new_cluster to `self.current_clusters` and `self.clusters_storage`.
        self.current_clusters.append(new_cluster_ID)
        self.clusters_storage[new_cluster_ID] = new_cluster

        return new_cluster_ID

    # ==============================================================================
    # ADD CLUSTER BY MERGING CLUSTERS :
    # ==============================================================================
    def _add_new_cluster_by_merging_clusters(
        self,
        children: List[int],
    ) -> int:
        """
        Create or Update a cluster by setting its children, and add it to the storage and current clusters.

        Args:
            children (List[int]): A list of cluster IDs to define the new cluster by its children.

        Returns:
            int : ID of the merged cluster.
        """

        # Remove all leaves children clusters from `self.current_clusters`.
        for child_ID_to_remove in children:
            self.current_clusters.remove(child_ID_to_remove)

        """
        ###
        ### Tree optimization : if only one node, then update this node as parent of all leaves.
        ### TODO : test of check if relevant to use. pros = smarter tree visualisation ; cons = cluster number more difficult to choose.
        ###

        # List of children nodes.
        list_of_children_nodes: List[int] = [
            child_ID
            for child_ID in children
            if len(self.clusters_storage[child_ID].children) > 0
        ]

        if len(list_of_children_nodes) == 1:

            # Get the ID of the cluster to update
            parent_cluster_ID: int = list_of_children_nodes[0]
            parent_cluster: Cluster = self.clusters_storage[parent_cluster_ID]

            # Add all leaves
            parent_cluster.add_new_children(
                new_children=[
                    self.clusters_storage[child_ID]
                    for child_ID in children
                    if child_ID != parent_cluster_ID
                ],
                new_clustering_iteration=self.clustering_iteration
            )

            # Add new_cluster to `self.current_clusters` and `self.clusters_storage`.
            self.current_clusters.append(parent_cluster_ID)
            self.clusters_storage[parent_cluster_ID] = parent_cluster

            # Return the cluster_ID of the created cluster.
            return parent_cluster_ID


        """

        ###
        ### Default case : Create a new node as parent of all children to merge.
        ###

        # Get the ID of the new cluster.
        parent_cluster_ID: int = max(self.clusters_storage) + 1

        # Create the cluster
        parent_cluster = Cluster(
            vectors=self.vectors,
            cluster_ID=parent_cluster_ID,
            clustering_iteration=self.clustering_iteration,
            children=[self.clusters_storage[child_ID] for child_ID in children],
        )

        # Add new_cluster to `self.current_clusters` and `self.clusters_storage`.
        self.current_clusters.append(parent_cluster_ID)
        self.clusters_storage[parent_cluster_ID] = parent_cluster

        # Return the cluster_ID of the created cluster.
        return parent_cluster_ID

    # ==============================================================================
    # COMPUTE DISTANCE BETWEEN CLUSTERINGNEW ITERATION OF CLUSTERING :
    # ==============================================================================
    def _compute_distance(self, cluster_IDi: int, cluster_IDj: int) -> float:
        """
        Compute distance between two clusters.

        Args:
            cluster_IDi (int): ID of the first cluster.
            cluster_IDj (int): ID of the second cluster.

        Returns:
            float : Distance between the two clusters.
        """

        # Check `"CANNOT_LINK"` constraints.
        for data_ID1 in self.clusters_storage[cluster_IDi].members:
            for data_ID2 in self.clusters_storage[cluster_IDj].members:
                if (
                    self.constraints_manager.get_inferred_constraint(
                        data_ID1=data_ID1,
                        data_ID2=data_ID2,
                    )
                    == "CANNOT_LINK"
                ):
                    return float("Inf")

        # Case 1 : `self.linkage` is "complete".
        if self.linkage == "complete":

            return max(
                [
                    self.pairwise_distances_matrix[data_ID_in_cluster_IDi][data_ID_in_cluster_IDj]
                    for data_ID_in_cluster_IDi in self.clusters_storage[cluster_IDi].members
                    for data_ID_in_cluster_IDj in self.clusters_storage[cluster_IDj].members
                ]
            )

        # Case 2 : `self.linkage` is "average".
        if self.linkage == "average":
            vector_centroid1 = self.clusters_storage[cluster_IDi].centroid
            vector_centroid2 = self.clusters_storage[cluster_IDj].centroid
            return pairwise_distances(
                X=vector_centroid1,
                Y=vector_centroid2,
                metric="euclidean",  # TODO: Load different parameters for distance computing ?
            )[0][0].astype(np.float64)

        # Case 3 : `self.linkage` is "single".
        if self.linkage == "single":
            return min(
                [
                    self.pairwise_distances_matrix[data_ID_in_cluster_IDi][data_ID_in_cluster_IDj]
                    for data_ID_in_cluster_IDi in self.clusters_storage[cluster_IDi].members
                    for data_ID_in_cluster_IDj in self.clusters_storage[cluster_IDj].members
                ]
            )

        # Case 4 : `self.linkage` is "ward".
        ##if self.linkage == "ward": ## DEFAULTS
        # Compute distance
        merged_members: List[str] = (
            self.clusters_storage[cluster_IDi].members + self.clusters_storage[cluster_IDj].members
        )
        return sum(
            [
                self.pairwise_distances_matrix[data_IDi][data_IDj]
                for i, data_IDi in enumerate(merged_members)
                for j, data_IDj in enumerate(merged_members)
                if i < j
            ]
        ) / (len(self.clusters_storage[cluster_IDi].members) * len(self.clusters_storage[cluster_IDj].members))

    # ==============================================================================
    # DISTANCE : GETTER
    # ==============================================================================
    def _get_distance(self, cluster_IDi: int, cluster_IDj: int) -> float:
        """
        Get the distance between two clusters.

        Args:
            cluster_IDi (int): ID of the first cluster.
            cluster_IDj (int): ID of the second cluster.

        Returns:
            float : Distance between the two clusters.
        """

        # Sort IDs of cluster.
        min_cluster_ID: int = min(cluster_IDi, cluster_IDj)
        max_cluster_ID: int = max(cluster_IDi, cluster_IDj)

        # Return the distance.
        return self.clusters_distance[min_cluster_ID][max_cluster_ID]

    # ==============================================================================
    # DISTANCE : SETTER
    # ==============================================================================
    def _set_distance(
        self,
        distance: float,
        cluster_IDi: int,
        cluster_IDj: int,
    ) -> None:
        """
        Set the distance between two clusters.

        Args:
            distance (float): The distance between the two clusters.
            cluster_IDi (int): ID of the first cluster.
            cluster_IDj (int): ID of the second cluster.
        """

        # Sort IDs of cluster.
        min_cluster_ID: int = min(cluster_IDi, cluster_IDj)
        max_cluster_ID: int = max(cluster_IDi, cluster_IDj)

        # Add distance to the dictionary of distance.
        if min_cluster_ID not in self.clusters_distance:
            self.clusters_distance[min_cluster_ID] = {}
        self.clusters_distance[min_cluster_ID][max_cluster_ID] = distance

    # ==============================================================================
    # GET THE TWO CLOSEST CLUSTERS
    # ==============================================================================
    def _get_the_two_clostest_clusters(self) -> Optional[Tuple[int, int]]:
        """
        Get the two clusters which are the two closest clusters.

        Returns:
            Optional(Tuple[int, int]) : The IDs of the two closest clusters to merge. Return None if no cluster is suitable.
        """

        # Compute the two clostest clusters to merge. take the closest distance, then the closest cluster size.
        clostest_clusters = min(
            [
                {
                    "cluster_ID1": cluster_ID1,
                    "cluster_ID2": cluster_ID2,
                    "distance": self._get_distance(cluster_IDi=cluster_ID1, cluster_IDj=cluster_ID2),
                    "merged_size": len(self.clusters_storage[cluster_ID1].members)
                    + len(self.clusters_storage[cluster_ID2].members)
                    # TODO : Choose between "distance then size(count)" and "size_type(boolean) then distance"
                }
                for cluster_ID1 in self.current_clusters
                for cluster_ID2 in self.current_clusters
                if cluster_ID1 < cluster_ID2
            ],
            key=lambda dst: (dst["distance"], dst["merged_size"]),
        )

        # Get clusters and distance.
        cluster_ID1: int = int(clostest_clusters["cluster_ID1"])
        cluster_ID2: int = int(clostest_clusters["cluster_ID2"])
        distance: float = clostest_clusters["distance"]

        # Check distance.
        if distance == float("Inf"):
            return None

        # Return the tow closest clusters.
        return cluster_ID1, cluster_ID2

    # ==============================================================================
    # COMPUTE PREDICTED CLUSTERS
    # ==============================================================================
    def compute_predicted_clusters(self, nb_clusters: int, by: str = "size") -> Dict[str, int]:
        """
        Compute the predicted clusters based on clustering tree and estimation of number of clusters.

        Args:
            nb_clusters (int): The number of clusters to compute.
            by (str, optional): A string to identifies the criteria used to explore `HierarchicalConstrainedClustering` tree. Can be `"size"` or `"iteration"`. Defaults to `"size"`.

        Raises:
            ValueError: if `clustering_root` was not set.

        Returns:
            Dict[str,int] : A dictionary that contains the predicted cluster for each data ID.
        """

        # Check that the clustering has been made.
        if self.clustering_root is None:
            raise ValueError("The `clustering_root` is not set, probably because clustering was not run.")

        ###
        ### EXPLORE CLUSTER TREE
        ###

        # Define the resulted list of children as the children of `HierarchicalConstrainedClustering` root.
        list_of_clusters: List[Cluster] = [self.clustering_root]

        # Explore `HierarchicalConstrainedClustering` children until dict_of_predicted_clusters has the right number of children.
        while len(list_of_clusters) < nb_clusters:

            if by == "size":
                # Get the biggest cluster in current children from `HierarchicalConstrainedClustering` exploration.
                # i.e. it's the cluster that has the more data to split.
                cluster_to_split = max(list_of_clusters, key=lambda c: len(c.members))

            else:  # if by == "iteration":
                # Get the most recent cluster in current children from `HierarchicalConstrainedClustering` exploration.
                # i.e. it's the cluster that was last merged.
                cluster_to_split = max(list_of_clusters, key=lambda c: c.clustering_iteration)

            # If the chosen cluster is a leaf : break the `HierarchicalConstrainedClustering` exploration.
            if cluster_to_split.children == []:  # noqa: WPS520
                break

            # Otherwise: The chosen cluster is a node, so split it and get its children.
            else:
                # ... remove the cluster obtained ...
                list_of_clusters.remove(cluster_to_split)

                # ... and add all its children.
                for child in cluster_to_split.children:
                    list_of_clusters.append(child)

        ###
        ### GET PREDICTED CLUSTERS
        ###

        # Initialize the dictionary of predicted clusters.
        predicted_clusters: Dict[str, int] = {data_ID: -1 for data_ID in self.list_of_data_IDs}

        # For all cluster...
        for cluster in list_of_clusters:

            # ... and for all member in each cluster...
            for data_ID in cluster.members:

                # ... affect the predicted cluster (cluster ID) to the data.
                predicted_clusters[data_ID] = cluster.cluster_ID

        # Rename cluster IDs by order.
        predicted_clusters = rename_clusters_by_order(clusters=predicted_clusters)

        # Return predicted clusters
        return predicted_clusters


# ==============================================================================
# CLUSTER
# ==============================================================================
class Cluster:
    """
    This class represents a cluster as a node of the hierarchical clustering tree.
    """

    # ==============================================================================
    # INITIALIZATION
    # ==============================================================================
    def __init__(
        self,
        vectors: Dict[str, Union[ndarray, csr_matrix]],
        cluster_ID: int,
        clustering_iteration: int,
        children: Optional[List["Cluster"]] = None,
        members: Optional[List[str]] = None,
    ) -> None:
        """
        The constructor for Cluster class.

        Args:
            vectors (Dict[str,Union[ndarray,csr_matrix]]): The representation of data vectors. The keys of the dictionary represents the data IDs. This keys have to refer to the list of data IDs managed by the `constraints_manager` (if `constraints_manager` is set). The value of the dictionary represent the vector of each data. Vectors can be dense (`numpy.ndarray`) or sparse (`scipy.sparse.csr_matrix`).
            cluster_ID (int): The cluster ID that is defined during `HierarchicalConstrainedClustering.cluster` running.
            clustering_iteration (int): The cluster iteration that is defined during `HierarchicalConstrainedClustering.cluster` running.
            children (Optional[List["Cluster"]], optional): A list of clusters children for cluster initialization. Incompatible with `members` parameter. Defaults to `None`.
            members (Optional[List[str]], optional): A list of data IDs for cluster initialization. Incompatible with `children` parameter. Defaults to `None`.

        Raises:
            ValueError: if `children` and `members` are both set or both unset.
        """

        # Store links to `vectors`.
        self.vectors: Dict[str, Union[ndarray, csr_matrix]] = vectors

        # Cluster ID and Clustering iteration.
        self.cluster_ID: int = cluster_ID
        self.clustering_iteration: int = clustering_iteration

        # Check children and members.
        if ((children is not None) and (members is not None)) or ((children is None) and (members is None)):
            raise ValueError(
                "Cluster initialization must be by `children` setting or by `members` setting, but not by both or none of them."
            )

        # Add children (empty or not).
        self.children: List["Cluster"] = children if (children is not None) else []

        # Cluster inverse depth.
        self.cluster_inverse_depth: int = (
            max([child.cluster_inverse_depth for child in self.children]) + 1 if (self.children) else 0
        )

        # Add members (empty or not).
        self.members: List[str] = (
            members if members is not None else [data_ID for child in self.children for data_ID in child.members]
        )

        # Update centroids
        self.update_centroid()

    # ==============================================================================
    # ADD NEW CHILDREN :
    # ==============================================================================
    def add_new_children(
        self,
        new_children: List["Cluster"],
        new_clustering_iteration: int,
    ) -> None:
        """
        Add new children to the cluster.

        Args:
            new_children (List["Cluster"]): The list of new clusters children to add.
            new_clustering_iteration (int): The new cluster iteration that is defined during HierarchicalConstrainedClustering.clusterize running.
        """

        # Update clustering iteration.
        self.clustering_iteration = new_clustering_iteration

        # Update children.
        self.children += [new_child for new_child in new_children if new_child not in self.children]

        # Update cluster inverse depth.
        self.cluster_inverse_depth = max([child.cluster_inverse_depth for child in self.children]) + 1

        # Update members.
        self.members = [data_ID for child in self.children for data_ID in child.members]

        # Update centroids.
        self.update_centroid()

    # ==============================================================================
    # UPDATE CENTROIDS :
    # ==============================================================================
    def update_centroid(self) -> None:
        """
        Update centroid of the cluster.
        """

        # Update centroids.
        self.centroid: Union[ndarray, csr_matrix] = (
            sum([self.vectors[data_ID] for data_ID in self.members]) / self.get_cluster_size()
        )

    # ==============================================================================
    # GET CLUSTER SIZE :
    # ==============================================================================
    def get_cluster_size(self) -> int:
        """
        Get cluster size.

        Returns:
            int: The cluster size, i.e. the number of members in the cluster.
        """

        # Update centroids.
        return len(self.members)

    # ==============================================================================
    # TO DICTIONARY :
    # ==============================================================================
    def to_dict(self) -> Dict[str, Any]:
        """
        Transform the Cluster object into a dictionary. It can be used before serialize this object in JSON.

        Returns:
            Dict[str, Any]: A dictionary that represents the Cluster object.
        """

        # Define the result dictionary.
        results: Dict[str, Any] = {}

        # Add clustering information.
        results["cluster_ID"] = self.cluster_ID
        results["clustering_iteration"] = self.clustering_iteration

        # Add children information.
        results["children"] = [child.to_dict() for child in self.children]
        results["cluster_inverse_depth"] = self.cluster_inverse_depth

        # Add members information.
        results["members"] = self.members

        return results
